'''
Library containing the main functions for the execution of the model.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''

import pandas as pd

class inputs_simulation (object):
    def __init__(self,inputs_cal):
            
        if type(inputs_cal) != str:
            self.params_cal = inputs_cal.Fitted_parameters
        elif type(inputs_cal) == str:
            self.params_cal = pd.read_csv(inputs_cal,index_col=0)
        else:
            raise Exception


'''
Library containing classes for calibrating model parameters.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''

from NEOPRENE.NSRP.MathematicalPropertiesNSRP import *
from NEOPRENE.NSRP.utils import *
from NEOPRENE.NSRP.libs_NSRP import *
import time
import yaml
import ast
from NEOPRENE.NSRP.parameters_calibration import parameters_calibration

class Calibration(object):
    """This function allows you to configure the point model to calibrate from user-defined hyperparameters.
    
    Parameters
    ----------
    hiperparams : Object. 
        Object with calibration hyperparameters.
      
    """
    def __init__(self,hiperparams):
        self.hiperparams = hiperparams
        
    def __call__(self, statistics, verbose=False):
        
        """Starting from the general hyperparameters of the model and different real statistics, the calibration of the point model begins.
        
        Parameters
        ----------
        statistics: Object. 
            Object with statistics.
          
        """
        
        
    
        self.statististics_dataframe   = statistics.statististics_dataframe
   
        ##Limits are obtained at the indicated time resolution
        if self.hiperparams.temporal_resolution=='d':
            t=24
        else:
            t=1
            
        lim = np.array([
            [(1/self.hiperparams.time_between_storms[1])*t, (1/self.hiperparams.time_between_storms[0])*t],
            [self.hiperparams.number_storm_cells[0], self.hiperparams.number_storm_cells[1]],
            [(1/self.hiperparams.cell_duration[1])*t, (1/self.hiperparams.cell_duration[0])*t],
            [(1/self.hiperparams.cell_intensity[1])*t, (1/self.hiperparams.cell_intensity[0])*t], 
            [(1/self.hiperparams.storm_cell_displacement[1])*t, (1/self.hiperparams.storm_cell_displacement[0])*t]])
            
        statististics_fit_dataframe=pd.DataFrame(index=self.hiperparams.statistics_name,columns=self.hiperparams.Seasonality_str)

        if self.hiperparams.process=='normal':
            param_s=['landa', 'ipsilon', 'eta', 'xi', 'betha'] 
            
            # We create a dataframe to store the adjusted intensity function parameters.
            Dataframe_params=pd.DataFrame(index=param_s,columns=self.hiperparams.Seasonality_str); n_p=5;
            limites=np.array([[lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],[lim[3,0],lim[3,1]],\
                              [lim[4,0],lim[4,1]]])

        elif self.hiperparams.process=='storms':
            param_s=['landa1', 'ipsilon1', 'eta1', 'xi1', 'betha1', 'landa2', 'ipsilon2', 'eta2', 'xi2', 'betha2']
            
            # We create a dataframe to store the adjusted intensity function parameters.
            Dataframe_params=pd.DataFrame(index=param_s,columns=self.hiperparams.Seasonality_str); n_p=10
            limites=np.array([[lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],[lim[3,0],lim[3,1]],\
                              [lim[4,0],lim[4,1]], [lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],\
                              [lim[3,0],lim[3,1]],[lim[4,0],lim[4,1]]]) 
                              
        ## We start the adjustment for each with seasonality.
        print('')
        print('')
        print('#'*80)
        print('Adjustment of parameters using the Particle Swarm Optimization (PSO)')
        print('')
        print('')
        Error_dataframe=pd.DataFrame(index=self.hiperparams.Seasonality_str)
        Error_dataframe['Total Error']=np.nan
        
        for pri, prii in enumerate(self.hiperparams.Seasonality):
            evaluator = evaluateInd_PSO(self.statististics_dataframe.loc[:,str(prii)].values, self.hiperparams.weights, self.hiperparams.process, self.hiperparams.statistics_name)
            if verbose:
                print('')
                print('Fitting the months = ' + str(prii))
                print('')
            error=0.5
            error_initializations=list()
            parameters_initializations=list()
            n_i=0
            total_error=10000
            while total_error>error and self.hiperparams.number_initializations>n_i:
                if verbose:
                    print('Initialization number  = ' + str(n_i))
                n_i+=1
                Steps = 0
                mySwarm = pso.Swarm(n_p, limites, self.hiperparams.number_bees, evaluator, 'min', verbose)
                
                while True:
                    if verbose:
                        print('Iteration number  = ' + str(Steps))
                    mySwarm.step()
                    Steps += 1
                    (b, bp) = mySwarm.getBest()
                    if verbose:
                        print('Total error = ' + str(evaluator.totalError(bp)))
                        if verbose:
                            if total_error <= error:
                                print('Stopped calibration since total error is equal or lower than error bound (' + str(error) + ')') 

                    if (evaluator.totalError(bp) < error): break
                    elif Steps > self.hiperparams.number_iterations: ###1000
                        break
                error_initializations.append(evaluator.totalError(bp))	
                parameters_initializations.append(bp)
                sta_fit=evaluator.compute(bp);
                statististics_fit_dataframe.loc[:,str(prii)]= sta_fit[0]
                Fitted_parameters=mySwarm.getBest()[1]
                Error_dataframe.loc[str(prii)]=evaluator.totalError(bp)

                total_error=evaluator.totalError(bp)

            if n_i>1:
                # If in the first initialisation it does not find the best parameters, it does it again and keeps the simulation with the lowest error.
                Error_dataframe.loc[str(prii)]=np.min(error_initializations)
                sta_fit=evaluator.compute(parameters_initializations[np.argmin(error_initializations)]);
                statististics_fit_dataframe.loc[:,str(prii)]= sta_fit[0]
                Fitted_parameters=parameters_initializations[np.argmin(error_initializations)]


            if self.hiperparams.process=='normal':

                param_v=list(); 
                param_v.append(Fitted_parameters[0]);
                param_v.append(Fitted_parameters[1]); 
                param_v.append(Fitted_parameters[2]); 
                param_v.append(Fitted_parameters[3]); 
                param_v.append(Fitted_parameters[4]); 
                Dataframe_params.loc[:,str(prii)]=param_v



            elif self.hiperparams.process=='storms':

                param_v=list();
                param_v.append(Fitted_parameters[0]);
                param_v.append(Fitted_parameters[1]); 
                param_v.append(Fitted_parameters[2]); 
                param_v.append(Fitted_parameters[3]); 
                param_v.append(Fitted_parameters[4]); 
                param_v.append(Fitted_parameters[5]); 
                param_v.append(Fitted_parameters[6]); 
                param_v.append(Fitted_parameters[7]); 
                param_v.append(Fitted_parameters[8]); 
                param_v.append(Fitted_parameters[9]); 
                Dataframe_params.loc[:,str(prii)]=param_v

        resuls = parameters_calibration(Dataframe_params,statististics_fit_dataframe,self.statististics_dataframe, Error_dataframe)
        
        
        return resuls
    



'''
Library containing classes for calibrating model parameters.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''

from NEOPRENE.NSRP.utils import *

def compare_statistics(CAL, SIM, frecuency):
    
    stats_obs=allmonths(CAL.statististics_Real)
    stats_fit=allmonths(CAL.statististics_Fit)
    stats_sim=allmonths(SIM.statististics_Simulated)
    
    N = len(stats_obs.index)
    RK = int(np.ceil(np.sqrt(len(stats_obs.index)))); 
    CK = int(math.ceil(N / RK))
        
    delete =  RK*CK-len(stats_obs.index)
           
    fig, axes = plt.subplots(RK, CK, figsize=(15, 15))
    
    axes = axes.ravel()
    for i, ii in enumerate(stats_obs.index):
        if 'mean' in ii:
            name_statistics = r'$\mu$'
        elif 'var' in ii:
            name_statistics = r'$\sigma_{'+ii.split('_')[-1]+frecuency+'}$'
        elif 'autocorr' in ii:
            name_statistics = r'$ACF-lag1_{'+ii.split('_')[-2]+frecuency+'}$'
        elif 'fih' in ii:
            name_statistics = r'$\phi_{'+ii.split('_')[-1]+frecuency+'}$'
        elif 'fiWW' in ii:
            name_statistics = r'$\phi^{WW}_{'+ii.split('_')[-1]+frecuency+'}$'
        elif 'fiDD' in ii:
            name_statistics = r'$\phi^{DD}_{'+ii.split('_')[-1]+frecuency+'}$'
        elif 'M3' in ii:
            #name_statistics = r'$\overline{\mu}_{3_'+ii.split('_')[-1]+frecuency+'}$'
            name_statistics = r'$\bar{\mu}_{3_{'+ii.split('_')[-1]+frecuency+'}}$'
        
        Data_sta=pd.DataFrame(index=np.arange(1, 13))
        Obs=list(); Fit=list(); Sim=list();
        for m in np.arange(1, 13):
            Obs.append(stats_obs[m].loc[ii])
            Fit.append(stats_fit[m][ii])
            Sim.append(stats_sim[m].loc[ii])
            
        Data_sta.index = ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec']

        Data_sta['Obs']=Obs
        Data_sta['Fit']=Fit
        Data_sta['Sim']=Sim
        #Ploteo
        ax=axes[i]
        ax.set_title(name_statistics,fontsize=18)
        ax.grid(True)
        legnd=['Obs', 'Fit', 'Sim']
        pp=Data_sta['Obs'].plot(style='k--', lw=2,  ax=axes[i])
        Data_sta['Fit'].plot(style='bs', ax=axes[i])
        Data_sta['Sim'].plot(style='r^', ax=axes[i])
        #pp.legend(legnd, loc='best', )
        if ii[0:2]==('fi' or 'au'):
            ax.set_ylim(0, 1)
            
        ax.set_xticks(np.arange(0,12))
        ax.set_xticklabels(['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec'],rotation = 45)
            
        ax.grid()
            
        del name_statistics     
    if delete!=0:
        for i in range(1,delete+1):
            axes[-i].remove()
        
    lines = []
    labels = []
    for i,ax in enumerate(fig.axes):
        axLine, axLabel = ax.get_legend_handles_labels()
        lines.extend(axLine)
        labels.extend(axLabel)
    fig.legend(lines[:3], labels[:3],           
               loc = 8,ncol=3,fontsize=15)
    fig.tight_layout(pad=3.7)
    
    return fig

def exceedence_probability(Serie_Observed, Serie_Simulated, temporal_resolution):
    # Observed
    aux = Serie_Observed.values
    aux = aux[aux>=0]
    s_obs = np.sort(aux, axis=None)[::-1]
    exc_obs = np.arange(1.,len(s_obs)+1) / len(s_obs)

    # Simulated
    if temporal_resolution == 'd':
        units = 'mm/day'
    elif temporal_resolution == 'h':
        units = 'mm/hour'

    aux = Serie_Simulated.values; aux=aux[aux>=0]
    s_sim = np.sort(aux, axis=None)[::-1]
    exc_sim = np.arange(1.,len(s_sim)+1) / len(s_sim)
    
    fig, ax = plt.subplots(figsize=(10, 5))

    ax.plot(exc_obs, s_obs, '.k', markersize=8,  label='Obs')
    ax.plot(exc_sim, s_sim, 'r',  label='Sim')
    ax.set_xlabel("Exceedence Probability", fontsize=13)
    ax.set_ylabel(units, fontsize=13)
    ax.set_xscale('log')
    ax.grid(True)
    ax.tick_params(labelsize=13)
    ax.legend(fontsize = 15)
    
    return fig

def disaggregation_rainfall(x_series, y_series):
    """
    Dissagregation function from: A spatial–temporal point process model of rainfall
    for the Thames catchment, UK (Cowpertwait 2005). Eq:15
    
    Inputs:
    x_series: Daily observed series (DataFrame)
    y_series: Hourly simulated series (DataFrame)
    
    Outputs
    results: x_series disaggregated from daily-to-hourly
    """

    #y_series_daily = y_series.resample('D').agg(pd.Series.sum, min_count=1)
    #results=x_series.resample('h').agg(pd.Series.sum, min_count=1)*np.nan
    
    y_series_daily = y_series.resample('D').agg(pd.Series.sum, min_count=1)
    dti = pd.date_range(start=x_series.index[0], end=x_series.index[-1] + timedelta(hours=23), freq="H")
    #results=pd.DataFrame(index = dti, columns = ['Rain'])
    results=pd.DataFrame(index = dti, columns = y_series.columns)

    for n, date in enumerate(x_series.index[1:]):
        x_j=x_series.loc[date].values       

        if np.nansum(x_j)==0:#day with no rain
            posi_d=np.where((results.index.year==date.year) & (results.index.month==date.month) & (results.index.day==date.day))[0]
            results.iloc[posi_d]=0
        else:
            x_j_1=x_series.loc[x_series.index[n]].values
            y_j_AUX=y_series_daily.iloc[1:]
            y_j_1_AUX=y_series_daily.iloc[0:-1]

            poss=np.sum(y_j_AUX[y_j_AUX.columns[np.where(x_j>0)[0]]]>0, axis=1)==len(np.where(x_j>0)[0])#days where observed and simulated rain is equal

            y_j=y_j_AUX.loc[poss].copy()
            y_j_1=y_j_1_AUX.iloc[np.where(poss==True)[0]].copy()        

            suma=np.nansum(((x_j-y_j.values)**2)+((x_j_1-y_j_1.values)**2), axis=1)

            posi=np.argmin(suma)
            posi_day=y_j.index[posi]
            rr_day=y_j.values[posi,:]

            posi_d=np.where((results.index.year==date.year) & (results.index.month==date.month) & (results.index.day==date.day))[0]
            posi_h_s=np.where((y_series.index.year==posi_day.year) & (y_series.index.month==posi_day.month)\
                              & (y_series.index.day==posi_day.day))[0]

            results.iloc[posi_d]=y_series.iloc[posi_h_s].values
            
    return results

def figure_disaggregation(hourly_disaggregation,daily_disaggregation,real_series,yearmin,yearmax):
    import matplotlib as mpl
    from matplotlib.lines import Line2D
    import matplotlib.patches as mpatches
    
    t1 = str(yearmin) + '-02-01'; t2 = str(yearmax) + '-02-28'
    
    hourly_disaggregation_c = hourly_disaggregation.loc[t1:t2].copy()
    daily_disaggregation_c  = daily_disaggregation.loc[t1:t2].copy()
    x_series_c              = real_series.loc[t1:t2].copy()
    
    hourly_disaggregation_tab = pd.DataFrame(index=hourly_disaggregation_c.index,columns=['mm','day','Hour','Par_Impar'])

    hourly_disaggregation_tab.iloc[:,0] = hourly_disaggregation_c.values
    hourly_disaggregation_tab.iloc[:,1] = hourly_disaggregation_c.index.day
    hourly_disaggregation_tab.iloc[:,2] = hourly_disaggregation_c.index.hour
    hourly_disaggregation_tab.iloc[:,3] = (hourly_disaggregation_tab.day.values % 2)
    
    fig, ax = plt.subplots(figsize=(20, 5))
    ax2 = ax.twinx()
   
    f    = x_series_c.iloc[:,0].plot(color = 'b', style= '--',  ax=ax,label = 'Obs. (daily)')
    h_im = (hourly_disaggregation_tab[(hourly_disaggregation_tab.Par_Impar==1)&((hourly_disaggregation_tab.mm>0.1))].mm.dropna()).plot(color = 'k', style = 's', markersize = 5, ax = ax2, label = 'Disagg. (hourly) day odd')
    h_p  = (hourly_disaggregation_tab[(hourly_disaggregation_tab.Par_Impar==0)&((hourly_disaggregation_tab.mm>0.1))].mm.dropna()).plot(color = 'k', style = 'x', markersize = 5, ax = ax2,label = 'Disagg. (hourly) day even')
    h_l  = ax2.plot(hourly_disaggregation_c.index,hourly_disaggregation_c.values, '--', c='grey', label = 'Disagg. (hourly)')

    l    = daily_disaggregation_c.iloc[:,0].plot(color = 'r', style = '^', markersize = 8, ax = ax, label = 'Disagg. (daily)')

    # ax.legend(['Obs. (daily)','Disagg. (daily)'], fontsize = 15)
    # ax2.legend(['Disagg. (hourly) day odd','Disagg. (hourly) day even','Disagg. (hourly)'], fontsize = 15, loc=2)
    ax.set_xlim(t1, t2)
    #ax2.set_xlim(t1, t2)
    ax.tick_params(axis = 'both', labelsize = 15)
    ax2.tick_params(axis = 'both', labelsize = 15)
    ax.set_yticks(np.arange(0,np.max(daily_disaggregation_c.values)+8,4))
    ax.set_ylim(0, np.max(daily_disaggregation_c.values)+8)
    ax2.set_ylim(0, (np.max(daily_disaggregation_c.values)+8)/2)
    ax.set_ylabel('mm/day', fontsize = 15);
    ax2.set_ylabel('mm/hour', fontsize = 15);
    ax.set_xticks(daily_disaggregation_c.index)
    ax.set_xticklabels(np.arange(1,len(daily_disaggregation_c)+1))
    ax.grid()
    ax2.grid()
    
    legend_elements = [Line2D([0], [0], label='Obs. (daily)', color='b',linestyle = '--'),
                   Line2D([], [], label='Disagg. (daily)', color='r',marker='^',linestyle='None'),
                   Line2D([0], [0], label='Disagg. (hourly) day odd', color='k',marker='s',linestyle = '--'),
                   Line2D([0], [0], label='Disagg. (hourly) day even', color='k',marker='x',linestyle = '--')]

    fig.legend(handles=legend_elements, loc = 8, ncol=5,fontsize=15)
    
    fig.tight_layout(pad=3.7)
    
    return fig


class Analysis(object):
    """"This function allows different analyses of the results obtained with respect to the calibration and actual data.
    Parameters
    ----------
    hiperparams : Object. 
        Object with calibration hyperparameters."""
    
    def __init__(self,CAL,SIM):
        self.CAL = CAL
        self.SIM = SIM
        self.figures = []
        self.names_figures = []
        
    def disaggregate_rainfall(self,x_series, y_series):
        self.hourly_disaggregation = disaggregation_rainfall(x_series,y_series)
        
    def compare_statistics_fig(self):
        self.figures.append(compare_statistics(self.CAL, self.SIM, self.SIM.temporal_resolution))
        self.names_figures.append('Statistical_comparison')
        
    def exceedence_probability_fig(self,Serie_Observed, Serie_Simulated):
        self.figures.append(exceedence_probability(Serie_Observed.mean(axis=1), Serie_Simulated.mean(axis=1), self.SIM.temporal_resolution))
        self.names_figures.append('Exceedence_probability')
        
    def figure_disaggregation_fig(self,hourly_disaggregation,daily_disaggregation,real_series,yearmin,yearmax):
        self.figures.append(figure_disaggregation(hourly_disaggregation,daily_disaggregation,real_series,yearmin,yearmax))
        self.names_figures.append('Serie_disaggregation')
        
    def save_figures(self,path):
        for i,fig in enumerate(self.figures):
            fig.savefig(path+self.names_figures[i]+'.png',bbox_inches='tight')
        
        
    




'''
Library containing classes for keeping the simulated model parameters.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''


import pandas as pd

class parameters_calibration(object):
    def __init__(self, Dataframe_parametros, statististics_fit_dataframe, statististics_Real, Error_dataframe):
        self.Fitted_parameters  = Dataframe_parametros
        self.statististics_Fit  = statististics_fit_dataframe
        self.statististics_Real = statististics_Real
        self.Error_dataframe    = Error_dataframe

    def save_files(self, path_output_files): 
        self.Fitted_parameters.to_csv(path_output_files+'Calibrated_parameters.csv')
        self.statististics_Real.to_csv(path_output_files+'statististics_real.csv')
        self.statististics_Fit.to_csv(path_output_files+'statististics_fit.csv')
            
        


"""Library containing the complementary mathematical functions necessary for the operation of the code.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
        
"""



import sys
import matplotlib.pyplot as plt

import pandas as pd
import numpy as np
import datetime
from datetime import timedelta
import math

def allmonths(dataframe):
    """We move the dataframe to 12 columns with 12 months"""
    
    dataframe_meses=pd.DataFrame(index=dataframe.index, columns=np.arange(1,13))
    if len(dataframe.columns)==12:
        dataframe_meses.loc[:,:] = dataframe.values
    else:
        for i,ii in enumerate(dataframe.columns):
            for j in list(map(int, ii[1:-1].split(','))):
                dataframe_meses.loc[:,j] = dataframe.loc[:,ii]

   
    return dataframe_meses


def datetime2matlabdn(dt):
    mdn = dt + timedelta(days = 366)
    frac_seconds = (dt-datetime.datetime(dt.year,dt.month,dt.day,0,0,0)).seconds\
    /(24.0 * 60.0 * 60.0)
    frac_microseconds = dt.microsecond / (24.0 * 60.0 * 60.0 * 1000000.0)
    return mdn.toordinal() + frac_seconds + frac_microseconds



def datetime2matlabdnJavi(pdt, dt):
    '''
    Function that passes from a vector of pandas to a list of time in ordinal format.
    dt = resolutión days/hours/minutes/seconds
    '''
    
    if dt=='days':
        dt_ordinal=1
    elif dt=='hours':
        dt_ordinal=1/24
    elif dt=='minutes':
        dt_ordinal=1/(24*60)
    elif dt=='seconds':
        dt_ordinal=1/(24*60*60)
        
    d1 = datetime.datetime(pdt.year[0], pdt.month[0], pdt.day[0])# start date
    tt_ordinal_aux=list()
    
    dtt=0
    for i in range(len(pdt)):
        if i ==0:
            #tt_ordinal_origen=(d1).toordinal()
            tt_ordinal_origen=datetime2matlabdn(d1)
            tt_ordinal_aux.append(tt_ordinal_origen)
        else:
            tt_ordinal_aux.append(tt_ordinal_origen+dtt)
            
        dtt=dtt+dt_ordinal
        
    return tt_ordinal_aux

def func(x, a, b, c):
    """Function to which the spatial correlation values are fitted"""

    return a * np.exp(-b * x) + c

def distancia_f(x1, y1, x2, y2):
    dist=((x1-x2)**2 + (y1-y2)**2)**0.5
    return dist
def haversine(lon1, lat1, lon2, lat2):
    from math import radians, cos, sin, asin, sqrt
    """
    Calculate the great circle distance between two points 
    on the earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians 
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula 
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a)) 
    km = 6367 * c
    return km

    model.params
    model.plot_summary()
    RETURN_PERIODS.return_periods=model.return_periods
    RETURN_PERIODS.return_values=model.return_values
    plt.show()

def IDW_f(xx, yy, zz, x, y, betha):
    '''
    xx, yy --> Station coordinates
    zz --> variable to be interpolated
    x, y --> point where you want to get the result
    betha--> eweight exponent. The higher the exponent, the more weight is assigned to nearby points.
    '''
    dist=list()
    for i in range(len(xx)):
        dist.append(haversine(xx[i], yy[i], x, y))
    dist=np.array(dist)
    weights=(dist**(-betha))/np.sum(dist**(-betha))
    
    return np.sum(zz*weights)


 

	




'''
Library containing classes for simulating time series from the model.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''



from NEOPRENE.NSRP.MathematicalPropertiesNSRP import *
from NEOPRENE.NSRP.utils import *
from NEOPRENE.NSRP.libs_NSRP import *
from NEOPRENE.NSRP.inputs_simulation import *
from NEOPRENE.NSRP.outputs_simulation import *
import time

class Simulation(object):
    def __init__(self,hiperparams):
        self.hiperparams = hiperparams
        
    def __call__(self, params_cal):
   
        statististics_sim_df=pd.DataFrame(index=self.hiperparams.statistics_name,columns=self.hiperparams.Seasonality_str)
        
        Df_params = inputs_simulation(params_cal).params_cal
        
        
        Df_params=allmonths(Df_params)
        print('')
        print('')
        print('#'*80)
        print('Synthetic simulation')
        print('')
        print('')
		## We start with synthetic simulation and statistical calculations.
        
        if self.hiperparams.process=='normal':
            [Df_sim_join_hour_, Df_sim_join_day_, Intensidad_cellss, Duracion_horas_cellss]=\
            NSRP_simulation(Df_params, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, self.hiperparams.process, self.hiperparams.Seasonality)
            print('Total cumulative rainfall - Analytical estimation = ' + "{:15.2f}".format(np.sum(Intensidad_cellss*Duracion_horas_cellss)))
            print('Total cumulative rainfall -             Simulated = ' + "{:15.2f}".format(np.sum(Df_sim_join_day_.values)))
            
        elif self.hiperparams.process=='storms':

            Df_params1=pd.DataFrame(index=['landa', 'ipsilon', 'eta', 'xi', 'betha'],columns=Df_params.columns)
            for i in Df_params.columns:
                Df_params1.loc[:,i]=Df_params.loc[:,i].values[0:5]

            Df_params2=pd.DataFrame(index=['landa', 'ipsilon', 'eta', 'xi', 'betha'],columns=Df_params.columns)
            for i in Df_params.columns:
                Df_params2.loc[:,i]=Df_params.loc[:,i].values[5:]

            [Df_sim_join_hour_1, Df_sim_join_day_1, Intensidad_cellss1, Duracion_horas_cellss1]=\
            NSRP_simulation(Df_params1, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, 'normal', self.hiperparams.Seasonality)

            [Df_sim_join_hour_2, Df_sim_join_day_2, Intensidad_cellss2, Duracion_horas_cellss2]=\
            NSRP_simulation(Df_params2, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, 'normal', self.hiperparams.Seasonality)

			##Combino los dos processos
            Df_sim_join_hour_=pd.DataFrame(index=Df_sim_join_hour_1.index)
            Df_sim_join_hour_['Rain']=Df_sim_join_hour_1.values+Df_sim_join_hour_2.values
            Df_sim_join_day_=pd.DataFrame(index=Df_sim_join_day_2.index)
            Df_sim_join_day_['Rain']=Df_sim_join_day_1.values+Df_sim_join_day_2.values

            print('Total cumulative rainfall - Analytical estimation - Storm 1 = ' + "{:15.2f}".format(np.sum(Intensidad_cellss1*Duracion_horas_cellss1))) 
            print('Total cumulative rainfall - Analytical estimation - Storm 2 = ' + "{:15.2f}".format(np.sum(Intensidad_cellss2*Duracion_horas_cellss2)))  

            print('Total cumulative rainfall - Analytical estimation = ' + \
            "{:15.2f}".format(np.sum(Intensidad_cellss1*Duracion_horas_cellss1)+np.sum(Intensidad_cellss2*Duracion_horas_cellss2)))
            print('Total cumulative rainfall -             Simulated = ' + "{:15.2f}".format(np.sum(Df_sim_join_day_.values)))    


        #statististics_dataframe=allmonths(statististics_dataframe)
        #print('')
        #print('')
        #print('#'*80)
        #print('Results Validation')
        #print('')
        #print('')
        
        ## Real, adjusted and simulated statistical calculations.

        for pri, prii in enumerate(self.hiperparams.Seasonality):
            if self.hiperparams.temporal_resolution == 'd':
                Data=Df_sim_join_day_.copy()
            elif self.hiperparams.temporal_resolution == 'h':
                Data=Df_sim_join_hour_.copy()
            Data[Data['Rain']<0]=np.nan
            Data[Data['Rain']<0]=np.nan

            if len(self.hiperparams.Seasonality)==12:
             
                Data['Seasonality']=Data['Rain']*np.nan
                pos=np.where(Data.index.month == prii); pos=pos[0]
                Data['Seasonality'].iloc[pos]=Data['Rain'].iloc[pos]
                Pluvio_GS = Data['Seasonality'][Data['Seasonality']>=0]
                Pluvio_GS[Pluvio_GS<0.001]=0
                Data=Pluvio_GS.astype(float)
                
            else:
  
                Data['Seasonality']=Data['Rain']*np.nan
                for i, ii in enumerate(prii):
                    pos=np.where(Data.index.month == ii); pos=pos[0]
                    Data['Seasonality'].iloc[pos]=Data['Rain'].iloc[pos]
                Pluvio_GS = Data['Seasonality'][Data['Seasonality']>=0]
                Pluvio_GS[Pluvio_GS<0.001]=0
                Data=Pluvio_GS.astype(float)



            statististics_values_synthetic = calculate_statistics(Data,self.hiperparams.statistics_name,self.hiperparams.temporal_resolution)
            statististics_sim_df.loc[:,str(prii)]=statististics_values_synthetic  

        #statististics_sim_df=allmonths(statististics_sim_df)
        
        results = outputs_simulation(Df_sim_join_day_,Df_sim_join_hour_,statististics_sim_df, self.hiperparams.temporal_resolution)

        return results
        
    # def save_files(self,results,path_output_files):
        
    #     results.Hourly_Simulation.to_csv(path_output_files+'Time_serie_hourly_simulated.csv')
    #     results.statististics_Simulated.to_csv(path_output_files+'statistic_hourly_simulated.csv')
    #     results.Daily_Simulation.to_csv(path_output_files+'Time_serie_day_simulated.csv')



'''
Library containing classes for computing statistics from time series.
The statistics are then used to calibrate the model parameters

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''

import sys
from NEOPRENE.STNSRP.libs_STNSRP import *
import numpy as np
import pandas as pd
import os


def statistics_from_file(statistics_name, Seasonality_str, Seasonality, temporal_resolution, files_folder):
    
    statistics_gauges={}
    
    crosscorr = list()
    for st in statistics_name:
        if 'crosscorr' in st:
            crosscorr.append(st)
            
    cross_corr_gauges={ccorr: {} for ccorr in crosscorr}
    
    
    files = os.listdir(files_folder)
    for sea in Seasonality:
        for file in files:
            if str(sea) in file and 'statististics' in file:
                stats = pd.read_csv(files_folder+file,index_col = 0)
                statistics_gauges[sea] = stats
                
    for cr in crosscorr:
        for sea in Seasonality:
            for file in files:
                if str(sea) in file and cr in file:
                    stats = pd.read_csv(files_folder+file,index_col = 0)
                    cross_corr_gauges[cr][sea] = stats
                    
    return statistics_gauges, cross_corr_gauges
    
def statistics_from_serie(statistics_name, Seasonality_str, Seasonality, temporal_resolution, time_series):
    
    statistics_name_no_cross = [sn for sn in statistics_name if not 'crosscorr' in sn]# delete cross corr from statistcis_name

    statististics_dataframe=pd.DataFrame(index=statistics_name_no_cross,columns=Seasonality_str)
    Datos_ = pd.DataFrame(index=time_series.index,columns = ['Rain'])
    Datos_.loc[:,'Rain'] = time_series.iloc[:,0].values
        
    for pri, prii in enumerate(Seasonality):
        Datos=Datos_.copy(); Datos[Datos['Rain']<0]=np.nan
             
        if len(Seasonality)==12:
            ## We select only the dates (months) for which I am going to calculate the statistics to be adjusted later.
            Datos['Seasonality']=Datos.loc[:,'Rain']*np.nan
            pos=np.where(Datos.index.month == prii); pos=pos[0]
            Datos['Seasonality'].iloc[pos]=Datos['Rain'][pos]
            Pluvio_GS = Datos['Seasonality'][Datos['Seasonality']>=0]
            Datos=Pluvio_GS
            
        else:
            ## We select only the dates (months) for which I am going to calculate the statistics to be adjusted later.
            Datos['Seasonality'] = Datos.loc[:,'Rain']*np.nan
            for i, ii in enumerate(prii):
                pos=np.where(Datos.index.month == ii); pos=pos[0]
                Datos['Seasonality'].iloc[pos]=Datos['Rain'][pos]
            Pluvio_GS = Datos['Seasonality'][Datos['Seasonality']>=0]
            Datos=Pluvio_GS
        
        
            ## We calculate the defined statistics to be adjusted and I enter them in a dataframe.
        
        statististics_values = calculate_statistics(Datos,statistics_name_no_cross, temporal_resolution)
        statististics_dataframe.loc[:,str(prii)]=statististics_values
    return statististics_dataframe

def statistics_from_several_series(statistics_name, Seasonality_str, Seasonality, temporal_resolution, time_series):
    statistics_gauges={}
    for ID in time_series.columns:
        time_serie = pd.DataFrame(index = time_series.index)
        time_serie['Rain'] = time_series[ID].values
        statistics_gauges[ID] = statistics_from_serie(statistics_name, Seasonality_str, Seasonality, temporal_resolution, time_serie)

    return statistics_gauges

def cross_corr(statistics_name, Seasonality, temporal_resolution, time_series, Attributes, func, coordinates):
    cross_corr_stationality={}
    if np.sum(['cross' in i for i in statistics_name])>=1:
        pos=np.where(['cross' in i for i in statistics_name]); pos=pos[0]
        for i, ii in enumerate(pos):
            cross_corr_s_aux, Distance_correlation =cross_corr_stationality_f(time_series, Seasonality, Attributes, func, coordinates, \
                                                          statistics_name[pos[i]], temporal_resolution)

            cross_corr_stationality[statistics_name[pos[i]]]=cross_corr_s_aux

    return cross_corr_stationality, Distance_correlation

def change_statistics_dic_order(statististics_dataframe, Seasonality):
    new_dic = {}
    keys = [k for k in statististics_dataframe.keys()]
    for sea in Seasonality:
        dtFr = pd.DataFrame(index = statististics_dataframe[keys[0]].index)
        for gauge in statististics_dataframe.keys():
            dtFr[gauge] = statististics_dataframe[gauge][str(sea)]#check
        new_dic[sea] = dtFr
    return new_dic


class Statistics (object):
    """ This function allows the computation of the point model's spline statistics and the spatal cross correlation for the calibration from different data sources. 
    The input data can be a file with the statistics or a time series from which the statistics are calculated.
    
    Parameters
    ----------
    time_series : Dataframe or None. 
        Precipitation time series with a given resolution. In the case where the series is not available and only the statistics are available, it is necessary to define this parameter as None

    attributes : Dataframe or None. 
        Attributes of the gauges (ID, Lon, Lat, elevation)
        
    file : string or None
        File containing the database containing the statistics previously defined in the input file containing the parameters for calibration.
        
    Outputs
    -------
    statististics_dataframe : Dataframe with calculated statistics. 
    crosscorr_dataframe : Dataframe with cross correlation.
    """
    def __init__(self, hiperparams, time_series=None, files_folder=None, attributes=None):
    
        statistics_name     = hiperparams.statistics_name
        Seasonality_str     = hiperparams.Seasonality_str
        Seasonality         = hiperparams.Seasonality
        temporal_resolution = hiperparams.temporal_resolution
        coordinates         = hiperparams.coordinates
            
        if type(time_series)!=type(None) and files_folder != None:
            raise Exception ('It is not possible to enter two types of data')
        if type(time_series)!=type(None) and type(attributes)!=type(None):
            if all(time_series.columns!=attributes.ID):
                raise Exception ('Lenght/name/order of time series and attributes does not match')
            else:
                #self.statististics_dataframe = statistics_from_serie(statistics_name,Seasonality_str,Seasonality,temporal_resolution, time_series)
                statististics_dataframe_dic = statistics_from_several_series(statistics_name,Seasonality_str,Seasonality,temporal_resolution, time_series)
                self.statistics_dataframe   = change_statistics_dic_order(statististics_dataframe_dic, Seasonality)
                self.crosscorr_dataframe, self.crosscorr_dist_dataframe     = cross_corr(statistics_name,Seasonality,temporal_resolution, time_series, attributes, func, coordinates)

        elif files_folder != None:
            [self.statististics_dataframe, self.crosscorr_dataframe] = statistics_from_file(statistics_name,Seasonality_str,Seasonality,temporal_resolution, files_folder)
        else:
            raise Exception


'''
Library containing classes for reading yaml configuration files for calibration and simulation.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''



import yaml
import ast
import sys
import numpy as np

class Calibration(object):
    """This function allows to obtain the hyperparameters to start the calibration of the point model and to check if all parameters are correct.
    
    Parameters
    ----------
    file_yml : String. 
        Directory where the Yml file with the configuration to perform the model calibration is located.
      
    """
    def __init__(self,file_yml):
        with open(file_yml) as file:
            hiper_params = yaml.load(file, Loader=yaml.FullLoader)
        self.Seasonality_type        = hiper_params['Seasonality_type']
        self.Seasonality_user        = hiper_params['Seasonality_user']
        self.temporal_resolution     = hiper_params['temporal_resolution']
        self.process                 = hiper_params['process']
        self.statistics_name         = hiper_params['statistics_name']
        self.weights                 = hiper_params['weights']
        self.number_iterations       = hiper_params['number_iterations']
        self.number_bees             = hiper_params['number_bees']
        self.number_initializations  = hiper_params['number_initializations']
        self.time_between_storms     = hiper_params['time_between_storms']
        self.number_storm_cells      = hiper_params['number_storm_cells']
        self.cell_duration           = hiper_params['cell_duration']
        self.cell_intensity          = hiper_params['cell_intensity']
        self.storm_cell_displacement = hiper_params['storm_cell_displacement']

        statistics_name_string = ['mean', 'var', 'autocorr','fih', 'fiWW', 'fiDD', 'M3']
        temporal_resolution_string = ['d', 'h']
        process_string = ['normal', 'storms']

        if self.process not in process_string:
            raise Exception ('The name of the hyperparameter <process> is not correct, check that the name is among those listed below: normal, process')
            sys.exit(1)  

        if self.temporal_resolution not in temporal_resolution_string:
            raise Exception ('The name of the hyperparameter <temporal_resolution> is not correct, check that the name is among those listed below: h, d')
            sys.exit(1) 

        if not len(self.time_between_storms) == 2:
            raise Exception ('The lenght of the hyperparameter <time_between_storms> must to be equal to 2')
            sys.exit(1) 

        if not len(self.number_storm_cells) == 2:
            raise Exception ('The lenght of the hyperparameter <number_storm_cells> must to be equal to 2')
            sys.exit(1) 

        if not len(self.cell_duration) == 2:
            raise Exception ('The lenght of the hyperparameter <cell_duration> must to be equal to 2')
            sys.exit(1) 

        if not len(self.cell_intensity) == 2:
            raise Exception ('The lenght of the hyperparameter <cell_intensity> must to be equal to 2')
            sys.exit(1) 

        if not len(self.storm_cell_displacement) == 2:
            raise Exception ('The lenght of the hyperparameter <storm_cell_displacement> must to be equal to 2')
            sys.exit(1) 

        for stn in self.statistics_name:
            if  not 'mean' in stn:
                if not '_' in stn:
                    raise Exception ('Lag (l) or aggregation level (h) is not included for the statistic <' + stn + '>. (e.g. ' + stn +'_1)')
                    sys.exit(1) 
                else:
                    if stn.split('_')[0] not in statistics_name_string:
                        raise Exception ('The name of the statistic <' + stn.split('_')[0] +  '> is not correct, check that the name is among those listed below: mean_h', 'var_h', 'autocorr_l_h','fih_h', 'fiWW_h', 'fiDD_h', 'M3_h')
                        sys.exit(1) 
                    else:
                        if stn.split('_')[0] in ['var','fih', 'fiWW', 'fiDD', 'M3']:
                            if len(stn.split('_'))==2:
                                if not str(stn.split('_')[1]).isnumeric():
                                    raise Exception ('Aggregation level (h) fot the statistic <' + stn.split('_')[0] +  '> has to be a integer (e.g. ' + stn +'_1)')
                                    sys.exit(1)
                            else:
                                raise Exception ('The statistic <' + stn.split('_')[0] +  '> only accept  the aggregation level (h) (e.g. (e.g. ' + stn.split('_')[0] +'_1)')
                                sys.exit(1)

                        elif stn.split('_')[0] == 'autocorr':
                            if not len(stn.split('_'))==3:
                                raise Exception ('The statistic <' + stn.split('_')[0] +  '> has to include the lag (l) and the aggregation level (h) (e.g. autocorr_1_1 (autocorr_l_h))')
                                sys.exit(1)
                            else:
                                if not str(stn.split('_')[1]).isnumeric():
                                    raise Exception ('Aggregation level (h) fot the statistic <' + stn.split('_')[0] +  '> has to be an integer')
                                    sys.exit(1)
                                if not str(stn.split('_')[1]).isnumeric():
                                    raise Exception ('Lag (l) fot the statistic <' + stn.split('_')[0] +  '> has to be an integer')
                                    sys.exit(1)
            else:
                if '_' in stn:
                    raise Exception ('The statistic <mean> should not includ any king of Lag (l) or aggregation level (h). The correct form is <mean>')
                    sys.exit(1)



        if not len(self.statistics_name)==len(self.weights):
            raise Exception ('Hyperparameters <statistics_name> and <weights> have to have the same lenght')
            sys.exit(1) 

        
        if self.Seasonality_type == 'annual':
            Seasonality=list()
            Seasonality.append((1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ,11, 12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
            
        elif self.Seasonality_type == 'seasonal':
            Seasonality=list()
            Seasonality.append((1, 2, 3))
            Seasonality.append((4, 5, 6))
            Seasonality.append((7, 8, 9))
            Seasonality.append((10, 11, 12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
            
        elif self.Seasonality_type == 'monthly':
            Seasonality=list()
            Seasonality.append((1))
            Seasonality.append((2))
            Seasonality.append((3))
            Seasonality.append((4))
            Seasonality.append((5))
            Seasonality.append((6))
            Seasonality.append((7))
            Seasonality.append((8))
            Seasonality.append((9))
            Seasonality.append((10))
            Seasonality.append((11))
            Seasonality.append((12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
                
        elif self.Seasonality_type == 'user_defined':
        
            Seasonality_str = hiper_params['Seasonality_user']
            
            Seasonality = list()
            for i in self.Seasonality_user:
                Seasonality.append(ast.literal_eval(i))
                
            if type(hiper_params['Seasonality_user'][0])!=str:
                raise Exception ('The definition of the months that make up the user-defined seasonality is incorrect, check that it is a string.')
                sys.exit(1)

            res_tuple = type(Seasonality[0]) is tuple
            if res_tuple == False:
                Seasonality_flat_sort = np.sort(Seasonality)
            else:
                Seasonality_flat_sort = np.sort([item for sublist in Seasonality for item in sublist])
            if not len(Seasonality_flat_sort)==12:
                raise Exception ('The longitud of the hyperparameter <Seasonality_user> have to be equal to 12.')
                sys.exit(1)
    
            else:
                if not all(Seasonality_flat_sort == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]):
                    raise Exception ('All the months of the year (from 1 to 12) should be included in the hyperparameter <Seasonality_user>')
                    sys.exit(1)
                
        else:
            raise Exception ('The name of the seasonality type is not correct, check that the name is among those listed below: annual, monthly, seasonal, user_defined')
            sys.exit(1)       
        self.Seasonality_str = Seasonality_str        
        self.Seasonality = Seasonality
        
        
            
        if self.temporal_resolution != 'h' and self.temporal_resolution != 'd':
            raise Exception ("The time resolution of the point model is poorly defined. It should be 'h' or 'd'.")
            sys.exit(1) 
        
        if self.process != 'storms' and self.process != 'normal':
            raise Exception ("The process type is wrongly defined, it should be 'storms' or 'normal'.")
            sys.exit(1) 
            
            
        statistic_list = ['mean_h', 'var_h', 'autocorr_l_h','fih_h', 'fiWW_h', 'fiDD_h', 'M3_h']
        
        for i in self.statistics_name:
            if i.split('_')[0]== 'autocorr':
                st =  i.split('_')[0]+'_'+'l'+'_'+'h'
                if st not in statistic_list:
                    raise Exception ('The '+i+' statistic has not been correctly defined.')
                    sys.exit(1) 
            else:
                st = i.split('_')[0]+'_'+'h'
                if st not in statistic_list:
                    raise Exception ('The '+i+' statistic has not been correctly defined.')
                    sys.exit(1) 
            
            
        
        
        
class Simulation(object): 
    """This function allows to obtain the hyperparameters to start the simulation of the point model and to check if all parameters are correct.
        
        Parameters
        ----------
        file_yml : String. 
            Directory where the Yml file with the configuration to perform the model simulation is located.
    """
    def __init__(self,file_yml_sim): 
        with open(file_yml_sim) as file:
            hiper_params = yaml.load(file, Loader=yaml.FullLoader)
        self.Seasonality_type        = hiper_params['Seasonality_type']
        self.Seasonality_user        = hiper_params['Seasonality_user']
        self.statistics_name         = hiper_params['statistics_name']
        self.process                 = hiper_params['process']
        self.temporal_resolution     = hiper_params['temporal_resolution']
        self.year_ini                = hiper_params['year_ini']
        self.year_fin                = hiper_params['year_fin']

        statistics_name_string = ['mean', 'var', 'autocorr','fih', 'fiWW', 'fiDD', 'M3']
        temporal_resolution_string = ['d', 'h']
        process_string = ['normal', 'storms']

        if self.process not in process_string:
            raise Exception ('The name of the hyperparameter <process> is not correct, check that the name is among those listed below: normal, process')
            sys.exit(1)  

        if self.temporal_resolution not in temporal_resolution_string:
            raise Exception ('The name of the hyperparameter <temporal_resolution> is not correct, check that the name is among those listed below: h, d')
            sys.exit(1) 

        for stn in self.statistics_name:
            if stn.split('_')[0] not in statistics_name_string:
                raise Exception ('The name of the statistic <' + stn.split('_')[0] +  '> is not correct, check that the name is among those listed below: mean_h', 'var_h', 'autocorr_l_h','fih_h', 'fiWW_h', 'fiDD_h', 'M3_h')
                sys.exit(1) 

        if isinstance(self.year_ini, str) or isinstance(self.year_fin, str):
            raise Exception ('The hyperparameter <year_fin> and <year_ini> have to be a integer not a string')
            sys.exit(1) 
        else:
            if str(self.year_ini).isnumeric() and str(self.year_fin).isnumeric():
                if not self.year_fin > self.year_ini:
                    raise Exception ('The hyperparameter <year_fin> has to be higher than the hyperparameter <year_ini>')
                    sys.exit(1) 
            else:
                raise Exception ('The hyperparameter <year_fin> and <year_ini> have to be a integer')
                sys.exit(1) 

        for stn in self.statistics_name:
            if  not 'mean' in stn:
                if not '_' in stn:
                    raise Exception ('Lag (l) or aggregation level (h) is not included for the statistic <' + stn + '>. (e.g. ' + stn +'_1)')
                    sys.exit(1) 
                else:
                    if stn.split('_')[0] not in statistics_name_string:
                        raise Exception ('The name of the statistic <' + stn.split('_')[0] +  '> is not correct, check that the name is among those listed below: mean_h', 'var_h', 'autocorr_l_h','fih_h', 'fiWW_h', 'fiDD_h', 'M3_h')
                        sys.exit(1) 
                    else:
                        if stn.split('_')[0] in ['var','fih', 'fiWW', 'fiDD', 'M3']:
                            if len(stn.split('_'))==2:
                                if not str(stn.split('_')[1]).isnumeric():
                                    raise Exception ('Aggregation level (h) fot the statistic <' + stn.split('_')[0] +  '> has to be a integer (e.g. ' + stn +'_1)')
                                    sys.exit(1)
                            else:
                                raise Exception ('The statistic <' + stn.split('_')[0] +  '> only accept  the aggregation level (h) (e.g. (e.g. ' + stn.split('_')[0] +'_1)')
                                sys.exit(1)

                        elif stn.split('_')[0] == 'autocorr':
                            if not len(stn.split('_'))==3:
                                raise Exception ('The statistic <' + stn.split('_')[0] +  '> has to include the lag (l) and the aggregation level (h) (e.g. autocorr_1_1 (autocorr_l_h))')
                                sys.exit(1)
                            else:
                                if not str(stn.split('_')[1]).isnumeric():
                                    raise Exception ('Aggregation level (h) fot the statistic <' + stn.split('_')[0] +  '> has to be an integer')
                                    sys.exit(1)
                                if not str(stn.split('_')[1]).isnumeric():
                                    raise Exception ('Lag (l) fot the statistic <' + stn.split('_')[0] +  '> has to be an integer')
                                    sys.exit(1)
            else:
                if '_' in stn:
                    raise Exception ('The statistic <mean> should not includ any king of Lag (l) or aggregation level (h). The correct form is <mean>')
                    sys.exit(1)
        
        if self.Seasonality_type=='annual':
            Seasonality=list()
            Seasonality.append((1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ,11, 12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
            
        elif self.Seasonality_type=='seasonal':
            Seasonality=list()
            Seasonality.append((1, 2, 3))
            Seasonality.append((4, 5, 6))
            Seasonality.append((7, 8, 9))
            Seasonality.append((10, 11, 12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
            
        elif self.Seasonality_type =='monthly':
            Seasonality=list()
            Seasonality.append((1))
            Seasonality.append((2))
            Seasonality.append((3))
            Seasonality.append((4))
            Seasonality.append((5))
            Seasonality.append((6))
            Seasonality.append((7))
            Seasonality.append((8))
            Seasonality.append((9))
            Seasonality.append((10))
            Seasonality.append((11))
            Seasonality.append((12))
            
            Seasonality_str = list()
            for sr in Seasonality:
                Seasonality_str.append(str(sr))
                
        elif self.Seasonality_type =='user_defined':
        
            Seasonality_str = hiper_params['Seasonality_user']
            
            Seasonality = list()
            for i in self.Seasonality_user:
                Seasonality.append(ast.literal_eval(i))
                
            if type(hiper_params['Seasonality_user'][0])!=str:
                raise Exception ('The definition of the months that make up the user-defined seasonality is incorrect, check that it is a string.')
                sys.exit(1) 

            res_tuple = type(Seasonality[0]) is tuple
            if res_tuple == False:
                Seasonality_flat_sort = np.sort(Seasonality)
            else:
                Seasonality_flat_sort = np.sort([item for sublist in Seasonality for item in sublist])
            if not len(Seasonality_flat_sort)==12:
                raise Exception ('The longitud of the hyperparameter <Seasonality_user> have to be equal to 12.')
                sys.exit(1)
    
            else:
                if not all(Seasonality_flat_sort == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]):
                    raise Exception ('All the months of the year (from 1 to 12) should be included in the hyperparameter <Seasonality_user>')
                    sys.exit(1)
                
        else:
            raise Exception ('The name of the seasonality type is not correct, check that the name is among those listed below: annual, monthly, seasonal, user_defined')
            sys.exit(1) 
                
        self.Seasonality_str = Seasonality_str        
        self.Seasonality = Seasonality


'''
Library containing the necessary functions to simulate the Neyman-Scott process. 
Rectangular Pulses (NSRP).

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''

'''Functions of the NSRP mode'''

from NEOPRENE.NSRP.MathematicalPropertiesNSRP import *
from NEOPRENE.NSRP.utils import *
import NEOPRENE.NSRP.PSOMdJ as pso
import numpy as np
import pandas as pd
import scipy as sp
import math as mt
import datetime
from scipy.interpolate import interp1d
from datetime import date
from datetime import timedelta
from scipy import stats


def calculate_statistics(Data,statistics,temporal_resolution):
    """ 
    Calculation of the statistics for a given station
    The statistics that can be calculated are:
    statistics=[mean, var_h, autocorr_l_h, fih_h, fiWW_h, fiDD_h, M3_h]
    (l=lag, h=aggregation levels)

    """

    statististics_values_real=list()
    if temporal_resolution=='d':
        t='D'
    elif temporal_resolution=='h':
        t='h'

    if np.sum(['mean' in i for i in statistics])>=1:
        statististics_values_real.append(np.nanmean(Data))
    if np.sum(['var' in i for i in statistics])>=1:
        pos=np.where(['var' in i for i in statistics]); pos=pos[0]
        for i, ii in enumerate(pos):
            h=int(statistics[ii].split("_",1)[1])
            aux=Data.resample(str(h) + t).agg(pd.Series.sum, min_count=1); 
            statististics_values_real.append(np.nanvar(aux))
    if np.sum(['autocorr' in i for i in statistics])>=1:
        pos=np.where(['autocorr' in i for i in statistics]); pos=pos[0]
        for i, ii in enumerate(pos):
            l=int(statistics[ii].split("_",3)[1])
            h=int(statistics[ii].split("_",3)[2])
            aux=Data.resample(str(h) + t).agg(pd.Series.sum, min_count=1); 
            Autocorrelation_aux=aux.autocorr(lag=l) 
            if np.size(Autocorrelation_aux)>1: Autocorrelation_aux=Autocorrelation_aux[0] 
            statististics_values_real.append(Autocorrelation_aux)
    if np.sum(['fih' in i for i in statistics])>=1:
        pos=np.where(['fih' in i for i in statistics]); pos=pos[0]
        for i, ii in enumerate(pos):
            h=int(statistics[ii].split("_",1)[1])
            statististics_values_real.append(fi_h(Data, h))
    if np.sum(['fiWW' in i for i in statistics])>=1:
        pos=np.where(['fiWW' in i for i in statistics]); pos=pos[0]
        for i, ii in enumerate(pos):
            h=int(statistics[ii].split("_",1)[1])
            statististics_values_real.append(fi_WW(Data, h))
    if np.sum(['fiDD' in i for i in statistics])>=1:
        pos=np.where(['fiDD' in i for i in statistics]); pos=pos[0]
        for i, ii in enumerate(pos):
            h=int(statistics[ii].split("_",1)[1])
            statististics_values_real.append(fi_DD(Data, h))
    if np.sum(['M3' in i for i in statistics])>=1:
        pos=np.where(['M3' in i for i in statistics]); pos=pos[0]
        for i, ii in enumerate(pos):
            h=int(statistics[ii].split("_",1)[1])
            aux=Data.resample(str(h) + t ).agg(pd.Series.sum, min_count=1);
            statististics_values_real.append(sp.stats.moment(aux, moment=3, nan_policy='omit'))
    
    return statististics_values_real



class evaluateInd_PSO(object):
    def __init__(self, vector, weights, process, statistics):
        self.v = vector
        self.w = weights
        self.p = process
        self.s = statistics
    def __call__(self, ind):
    
        landa=list(); mu_c=list(); eta=list(); xi=list(); betha=list(); alpha=list(); alpha_p=list()
        if self.p=='normal':
            landa.append(ind[0]); mu_c.append(ind[1]); eta.append(ind[2]); xi.append(ind[3]); betha.append(ind[4]); alpha.append(1); alpha_p.append(1); 

        elif self.p=='storms':
            landa.append(ind[0]); mu_c.append(ind[1]); eta.append(ind[2]); xi.append(ind[3]); betha.append(ind[4]); alpha.append(1); alpha_p.append(1); 
            landa.append(ind[5]); mu_c.append(ind[6]); eta.append(ind[7]); xi.append(ind[8]); betha.append(ind[9]); alpha.append(1); alpha_p.append(1); 
            
        ##Mean
        if np.sum(['mean' in i for i in self.s])>=1:
            a = NSRP_mean(1, landa, mu_c, eta, xi, alpha, alpha_p)
            a = self.w[0]*((1-(self.v[0]/a))**2 + (1-(a/self.v[0]))**2); d_e={'e0':a}  

            
        ##Variance
        if np.sum(['var' in i for i in self.s])>=1:
            pos=np.where(['var' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_covariance(h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
        
        ##Autocorrlation
        if np.sum(['autocorr' in i for i in self.s])>=1:
            pos=np.where(['autocorr' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                l=int(self.s[ii].split("_",3)[1])
                h=int(self.s[ii].split("_",3)[2])
                a=NSRP_covariance(h,l, landa, mu_c, eta, xi, betha, alpha, alpha_p)\
                    /NSRP_covariance(h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a


        ##fi_h
        if np.sum(['fih' in i for i in self.s])>=1:
            pos=np.where(['fih' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a=NSRP_pdry(h, landa, mu_c, eta, betha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
        ##fi_WW
        if np.sum(['fiWW' in i for i in self.s])>=1:
            pos=np.where(['fiWW' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_fi_WW(h, landa, mu_c, eta, betha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
        ##fi_DD
        if np.sum(['fiDD' in i for i in self.s])>=1:
            pos=np.where(['fiDD' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_fi_DD(h, landa, mu_c, eta, betha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
        ##M3
        if np.sum(['M3' in i for i in self.s])>=1:
            pos=np.where(['M3' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_moments_order_3('Poisson',h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)

                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
        
        for k in d_e.keys():
            if isnan(d_e[k]): d_e[k] = 10000

        error=list(); error.append([d_e['e' + str(i)] for i in range(len(d_e.keys()))])
        
        return np.sum(error)

    def totalError(self, ind):

        landa=list(); mu_c=list(); eta=list(); xi=list(); betha=list(); alpha=list(); alpha_p=list()
        if self.p=='normal':
            landa.append(ind[0]); mu_c.append(ind[1]); eta.append(ind[2]); xi.append(ind[3]); betha.append(ind[4]); alpha.append(1); alpha_p.append(1); 
        elif self.p=='storms':
            landa.append(ind[0]); mu_c.append(ind[1]); eta.append(ind[2]); xi.append(ind[3]); betha.append(ind[4]); alpha.append(1); alpha_p.append(1); 
            landa.append(ind[5]); mu_c.append(ind[6]); eta.append(ind[7]); xi.append(ind[8]); betha.append(ind[9]); alpha.append(1); alpha_p.append(1); 
        
        ##Mean
        if np.sum(['mean' in i for i in self.s])>=1:
            a = NSRP_mean(1, landa, mu_c, eta, xi, alpha, alpha_p)
            a = self.w[0]*((1-(self.v[0]/a))**2 + (1-(a/self.v[0]))**2); d_e={'e0':a}
            
        ##Variance
        if np.sum(['var' in i for i in self.s])>=1:
            pos=np.where(['var' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_covariance(h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
                
        ##Autocorrelation
        if np.sum(['autocorr' in i for i in self.s])>=1:
            pos=np.where(['autocorr' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                l=int(self.s[ii].split("_",3)[1])
                h=int(self.s[ii].split("_",3)[2])
                a=NSRP_covariance(h,l, landa, mu_c, eta, xi, betha, alpha, alpha_p)\
                    /NSRP_covariance(h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
                
        ##fi_h
        if np.sum(['fih' in i for i in self.s])>=1:
            pos=np.where(['fih' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a=NSRP_pdry(h, landa, mu_c, eta, betha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
                
        ##fi_WW
        if np.sum(['fiWW' in i for i in self.s])>=1:
            pos=np.where(['fiWW' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_fi_WW(h, landa, mu_c, eta, betha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
                
        ##fi_DD
        if np.sum(['fiDD' in i for i in self.s])>=1:
            pos=np.where(['fiDD' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_fi_DD(h, landa, mu_c, eta, betha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
                
        ##M3
        if np.sum(['M3' in i for i in self.s])>=1:
            pos=np.where(['M3' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a=NSRP_moments_order_3('Poisson',h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)
                a = self.w[ii]*((1-(self.v[ii]/a))**2 + (1-(a/self.v[ii]))**2); d_e['e' + str(ii)]=a
        
        for k in d_e.keys():
            if isnan(d_e[k]): d_e[k] = 10000
                
        error=list(); error.append([d_e['e' + str(i)] for i in range(len(d_e.keys()))])
        
        return np.sum(error)
        
    def compute(self, ind):
        landa=list(); mu_c=list(); eta=list(); xi=list(); betha=list(); alpha=list(); alpha_p=list()
        if self.p=='normal':
            landa.append(ind[0]); mu_c.append(ind[1]); eta.append(ind[2]); xi.append(ind[3]); betha.append(ind[4]); alpha.append(1); alpha_p.append(1); 
        elif self.p=='storms':
            landa.append(ind[0]); mu_c.append(ind[1]); eta.append(ind[2]); xi.append(ind[3]); betha.append(ind[4]); alpha.append(1); alpha_p.append(1); 
            landa.append(ind[5]); mu_c.append(ind[6]); eta.append(ind[7]); xi.append(ind[8]); betha.append(ind[9]); alpha.append(1); alpha_p.append(1); 
        
        ##Mean
        if np.sum(['mean' in i for i in self.s])>=1:
            a = NSRP_mean(1, landa, mu_c, eta, xi, alpha, alpha_p)
            v={'v0':a}
        ##Variance
        if np.sum(['var' in i for i in self.s])>=1:
            pos=np.where(['var' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_covariance(h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)
                v['v' + str(ii)]=a
        ##Autocorrelation
        if np.sum(['autocorr' in i for i in self.s])>=1:
            pos=np.where(['autocorr' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                l=int(self.s[ii].split("_",3)[1])
                h=int(self.s[ii].split("_",3)[2])
                a=NSRP_covariance(h,l, landa, mu_c, eta, xi, betha, alpha, alpha_p)\
                    /NSRP_covariance(h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)
                v['v' + str(ii)]=a
        ##fi_h
        if np.sum(['fih' in i for i in self.s])>=1:
            pos=np.where(['fih' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a=NSRP_pdry(h, landa, mu_c, eta, betha, alpha_p)
                v['v' + str(ii)]=a
        ##fi_WW
        if np.sum(['fiWW' in i for i in self.s])>=1:
            pos=np.where(['fiWW' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_fi_WW(h, landa, mu_c, eta, betha, alpha_p)
                v['v' + str(ii)]=a
        ##fi_DD
        if np.sum(['fiDD' in i for i in self.s])>=1:
            pos=np.where(['fiDD' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a = NSRP_fi_DD(h, landa, mu_c, eta, betha, alpha_p)
                v['v' + str(ii)]=a
        ##M3
        if np.sum(['M3' in i for i in self.s])>=1:
            pos=np.where(['M3' in i for i in self.s]); pos=pos[0]
            for i, ii in enumerate(pos):
                h=int(self.s[ii].split("_",1)[1])
                a=NSRP_moments_order_3('Poisson',h,0, landa, mu_c, eta, xi, betha, alpha, alpha_p)
                v['v' + str(ii)]=a
            
        values=list(); values.append([v['v' + str(i)] for i in range(len(v.keys()))])
        
        return values
	
def NSRP_simulation(Params_month, year_ini, year_fin, temporal_resolution,process, Seasonality):

    landa=Params_month[Params_month.index=='landa'].values[0];Storm_origin=1/landa
    ipsilon=Params_month[Params_month.index=='ipsilon'].values[0];Number_cells_per_storm=ipsilon-1
    
    if process=='normal':
        eta=Params_month[Params_month.index=='eta'].values[0]; Time_cell=1/eta
        xi=Params_month[Params_month.index=='xi'].values[0]; Intensity_cell=1/xi

    betha=Params_month[Params_month.index=='betha'].values[0]; Dist_cell_origin=1/betha


    time_d=pd.period_range(start=str((year_ini)),  end=str((year_fin)), freq='D')
    time_h=pd.period_range(start=str((year_ini)),  end=str((year_fin)), freq='h')
    time_min=pd.period_range(start=str((year_ini)),  end=str((year_fin)), freq='min')

    Df_sim_join_day  =pd.DataFrame(index=time_d, columns = ['Rain'])
    Df_sim_join_hour =pd.DataFrame(index=time_h, columns = ['Rain'])


    Intensity_cells_total=list()
    Time_cells_total = list()

    for perio, monthss in enumerate(Seasonality):
        n_days=np.sum(np.in1d(Df_sim_join_day.index.month, monthss))
        position_day=np.in1d(Df_sim_join_day.index.month, monthss)
        position_hour=np.in1d(Df_sim_join_hour.index.month, monthss)

        Df_sim_day_aux  = pd.period_range('1800', periods=n_days, freq='D')
        Df_sim_hour_aux = pd.period_range('1800', periods=n_days*24, freq='h')
        Df_sim_min_aux  = pd.period_range('1800', periods=n_days*24*60, freq='min')

        time_star=datetime.datetime.strptime(str(Df_sim_day_aux[0]), '%Y-%m-%d')
        time_end=datetime.datetime.strptime(str(Df_sim_day_aux[-1]), '%Y-%m-%d')
        time_lapso=time_star
        
        time_storm_origins=list()
        n=0

        if np.size(monthss)==1:
            monthss=[monthss]
        else:
            monthss=[perio+1]


        while time_lapso < time_end:
            
            s = np.random.exponential(Storm_origin[monthss[0]-1], 1)
            if temporal_resolution=='d':
                time_lapso=time_lapso +  datetime.timedelta(days=s[0])
            elif temporal_resolution=='h':
                time_lapso=time_lapso +  datetime.timedelta(hours=s[0])

            time_storm_origins.append(time_lapso)
            n=n+1


        n_storms=len(time_storm_origins)
        
        Number_cell_per_storm=list()
        for i, ii in enumerate(time_storm_origins):
            Number_cell_per_storm.append(1 + (np.random.poisson(Number_cells_per_storm[monthss[0]-1], 1)))
        

        time0 = time_star
        time_ini_cells=list()
        time_fin_cells=list()
        Intensity_cells = list()
        Time_hours_cells = list()
        for i in range(n_storms):
            time1=time_storm_origins[i]
            Distance_hours_cell_sim=np.random.exponential(scale=Dist_cell_origin[monthss[0]-1], size=Number_cell_per_storm[i])

            if process=='normal':
                Time_hours_cell_sim=np.random.exponential(scale=Time_cell[monthss[0]-1], size=Number_cell_per_storm[i])
                
                Intensity_cell_sim=np.random.exponential(scale=Intensity_cell[monthss[0]-1], size=Number_cell_per_storm[i])


              

            for j in range(Number_cell_per_storm[i][0]):

                if temporal_resolution=='d':
                    time1_cell=time1 + datetime.timedelta(days=Distance_hours_cell_sim[j])
                    time2_cell=time1_cell + datetime.timedelta(days=Time_hours_cell_sim[j])

                elif temporal_resolution=='h':
                    time1_cell=time1 + datetime.timedelta(hours=Distance_hours_cell_sim[j])
                    time2_cell=time1_cell + datetime.timedelta(hours=Time_hours_cell_sim[j])

                Intensity=Intensity_cell_sim[j]
                time_ini_cells.append(time1_cell)
                time_fin_cells.append(time2_cell)
                Intensity_cells.append(Intensity)
                Time_hours_cells.append(Time_hours_cell_sim[j])

        ################################################################################
        time_ini_cells=np.array(time_ini_cells)
        time_fin_cells=np.array(time_fin_cells)
        Intensity_cells=np.array(Intensity_cells)
        Time_hours_cells=np.array(Time_hours_cells)
        #################################################################################
        ############################################################################
        tt=pd.period_range(start=Df_sim_day_aux[0],end=Df_sim_day_aux[-1], freq='min')
        tt_ordinal=tt.astype(np.int32)*60*10**9
        ############################################################################
        Anhos=list()
        for i, ii in enumerate(time_fin_cells):
            aux=ii; year_aux=aux.year
            Anhos.append(year_aux)
            
        In_datess=np.where(np.array(Anhos)<=Df_sim_day_aux.year[-1])
        
        time_ini_cellss=time_ini_cells[In_datess[0]]
        time_fin_cellss=time_fin_cells[In_datess[0]]
        
        Intensity_cellss=Intensity_cells[In_datess[0]]; Intensity_cells_total.append(Intensity_cellss)
        Time_hours_cellss=Time_hours_cells[In_datess[0]]; Time_cells_total.append(Time_hours_cellss)
        
        t_ini = np.hstack([time_ini_cellss, time_fin_cellss])
        if process=='cells':
            i_ini = np.hstack([Intensity_cellss.T, -Intensity_cellss.T])
            i_ini = i_ini[0]
        else: 
            i_ini = np.hstack([Intensity_cellss.T, -Intensity_cellss.T])
        orden = np.argsort(t_ini)
        t = t_ini[orden]
        i = np.cumsum(i_ini[orden])
        i[i<0] = 0
        rain=i.copy()
        t_ordinal = pd.PeriodIndex(t.astype(str),freq='N').astype(np.int32)
        if np.size(t_ordinal)==0:
            a=1
        else:
            rainfall = interp1d(t_ordinal, rain, kind="zero", bounds_error=False, fill_value=0.)
            rr = rainfall(tt_ordinal)
            Date=pd.DataFrame(index=tt)
            if temporal_resolution=='d':
                Date['Rain']=rr/(60*24)
            elif temporal_resolution=='h':
                Date['Rain']=rr/(60)
            df = Date.copy()
            del Date
            
            df2 = pd.DataFrame(df.groupby([df.index.year,df.index.month,df.index.day])[['Rain']].sum().values,index = pd.period_range(start=df.index[0],end=df.index[-1],freq='D'))
            df3 = pd.DataFrame(df.groupby([df.index.year,df.index.month,df.index.day,df.index.hour])[['Rain']].sum().values,index = pd.period_range(start=df.index[0],end=df.index[-1],freq='h'))


            Df_sim_join_day['Rain'].iloc[np.where(position_day)[0][0:len(df2)]]=df2.values.flatten()
            
            Df_sim_join_hour['Rain'].iloc[np.where(position_hour)[0][0:len(df3)]]=df3.values.flatten()
            
        del rr, rainfall, tt, t_ordinal,tt_ordinal, rain, i, t, orden, t_ini, i_ini,df, df3, df2
    return Df_sim_join_hour, Df_sim_join_day,\
           np.hstack(Intensity_cells_total), np.hstack(Time_cells_total)









'''
Library containing classes for calibrating model parameters.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''

from NEOPRENE.STNSRP.MathematicalPropertiesSTNSRP import *
from NEOPRENE.STNSRP.utils import *
from NEOPRENE.STNSRP.libs_STNSRP import *
import time
import yaml
import ast
from NEOPRENE.STNSRP.parameters_calibration import parameters_calibration
import sys

class Calibration(object):
    """This function allows you to configure the STNSRPM to calibrate from user-defined hyperparameters.
    
    Parameters
    ----------
    hiperparams : Object. 
        Object with calibration hyperparameters.
      
    """
    def __init__(self,hiperparams):
        self.hiperparams  = hiperparams
    def __call__(self, statistics, time_series, verbose=False):
        """Starting from the general hyperparameters of the model and different real statistics, the calibration of the point model begins.
        
        Parameters
        ----------
        statistics: Object. 
            Object with statistics.
          
        """
        #self.Input_Series = Input_Series.copy()
        
    
        self.statistics_dataframe     = statistics.statistics_dataframe.copy()
        self.crosscorr_dataframe      = statistics.crosscorr_dataframe.copy()
        self.crosscorr_dataframe_dist = statistics.crosscorr_dist_dataframe.copy()
   
        ##Limits are obtained at the indicated time resolution
        if self.hiperparams.temporal_resolution=='d':
            t=24
        else:
            t=1
            
        if self.hiperparams.storm_radius==True:
            
            lim = np.array([
                [(1/self.hiperparams.time_between_storms[1])*t, (1/self.hiperparams.time_between_storms[0])*t],
                [self.hiperparams.number_storm_cells[0], self.hiperparams.number_storm_cells[1]],
                [(1/self.hiperparams.cell_duration[1])*t, (1/self.hiperparams.cell_duration[0])*t],
                #[(1/self.hiperparams.cell_intensity[1])*t, (1/self.hiperparams.cell_intensity[0])*t], 
                [(1/self.hiperparams.storm_cell_displacement[1])*t, (1/self.hiperparams.storm_cell_displacement[0])*t],
                [(1/self.hiperparams.cell_radius[1]), (1/self.hiperparams.cell_radius[0])],
                [(1/self.hiperparams.storm_radius_p[1]), (1/self.hiperparams.storm_radius_p[0])]])
            
        elif self.hiperparams.storm_radius==False:
            
              lim = np.array([
                [(1/self.hiperparams.time_between_storms[1])*t, (1/self.hiperparams.time_between_storms[0])*t],
                [self.hiperparams.number_storm_cells[0], self.hiperparams.number_storm_cells[1]],
                [(1/self.hiperparams.cell_duration[1])*t, (1/self.hiperparams.cell_duration[0])*t],
                #[(1/self.hiperparams.cell_intensity[1])*t, (1/self.hiperparams.cell_intensity[0])*t], 
                [(1/self.hiperparams.storm_cell_displacement[1])*t, (1/self.hiperparams.storm_cell_displacement[0])*t],
                [(1/self.hiperparams.cell_radius[1]), (1/self.hiperparams.cell_radius[0])]])
            
            
        #statististics_fit_dataframe=pd.DataFrame(index=self.hiperparams.statistics_name,columns=self.hiperparams.Seasonality_str)
        sta_len = len(self.hiperparams.statistics_name)
        nn_cross = np.sum(['cross' in i for i in self.hiperparams.statistics_name])
        cross_corr_division = 10#
        statistics_dataframe_fit=pd.DataFrame(index = self.hiperparams.statistics_name[0:sta_len-nn_cross])
        #crosscorr_dataframe_fit=self.crosscorr_dataframe
        # create empty dictionary where overwritte the fitted results
        crosscorr_dataframe_fit = {}
        for key in self.crosscorr_dataframe.keys():
            crosscorr_dataframe_season = {}
            for key_season in self.crosscorr_dataframe[key].keys():
                dataframe_aux = pd.DataFrame(index = self.crosscorr_dataframe[key][key_season].index, columns = self.crosscorr_dataframe[key][key_season].columns)
                dataframe_aux['dist'] = self.crosscorr_dataframe[key][key_season]['dist']
                crosscorr_dataframe_season[key_season] = dataframe_aux
            crosscorr_dataframe_fit[key] = crosscorr_dataframe_season

        if self.hiperparams.process=='normal':

            if self.hiperparams.storm_radius==False:

                param_s=['landa', 'ipsilon', 'eta', 'betha', 'fi_may'] 
                
                # We create a dataframe to store the adjusted intensity function parameters.
                Dataframe_params=pd.DataFrame(index=param_s); n_p=5;
                limites=np.array([[lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],[lim[3,0],lim[3,1]],\
                                  [lim[4,0],lim[4,1]]])

            elif self.hiperparams.storm_radius==True:

                param_s=['landa', 'ipsilon', 'eta', 'betha', 'fi_may', 'fi_may_s'] 
                
                # We create a dataframe to store the adjusted intensity function parameters.
                Dataframe_params=pd.DataFrame(index=param_s); n_p=6;
                limites=np.array([[lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],[lim[3,0],lim[3,1]],\
                                  [lim[4,0],lim[4,1]], [lim[5,0],lim[5,1]]])

        elif self.hiperparams.process=='storms':

            if self.hiperparams.storm_radius==False:

                param_s=['landa1', 'ipsilon1', 'eta1', 'betha1', 'fi_may1', 'landa2', 'ipsilon2', 'eta2', 'betha2', 'fi_may2']
                
                # We create a dataframe to store the adjusted intensity function parameters.
                Dataframe_params=pd.DataFrame(index=param_s); n_p=10
                limites=np.array([[lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],[lim[3,0],lim[3,1]],\
                                  [lim[4,0],lim[4,1]],\
                                  [lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],[lim[3,0],lim[3,1]],\
                                  [lim[4,0],lim[4,1]]])

            elif self.hiperparams.storm_radius==True:

                param_s=['landa1', 'ipsilon1', 'eta1', 'betha1', 'fi_may1', 'fi_may_s1', 'landa2', 'ipsilon2', 'eta2', 'betha2', 'fi_may1', 'fi_may_s2']
                
                # We create a dataframe to store the adjusted intensity function parameters.
                Dataframe_params=pd.DataFrame(index=param_s); n_p=12
                limites=np.array([[lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],[lim[3,0],lim[3,1]],\
                                  [lim[4,0],lim[4,1]],[lim[5,0],lim[5,1]],\
                                  [lim[0,0],lim[0,1]],[lim[1,0],lim[1,1]],[lim[2,0],lim[2,1]],[lim[3,0],lim[3,1]],\
                                  [lim[4,0],lim[4,1]],[lim[5,0],lim[5,1]]]) 
                              
        ## We start the adjustment for each with seasonality.
        print('')
        print('')
        print('#'*80)
        print('Adjustment of parameters using the Particle Swarm Optimization (PSO)')
        print('')
        print('')
        
        Error_dataframe=pd.DataFrame(index=self.hiperparams.Seasonality_str)
        Error_dataframe['Total Error']=np.nan

        statistics_observed_dataframe=pd.DataFrame(index=self.hiperparams.statistics_name)

        
        for pri, prii in enumerate(self.hiperparams.Seasonality):
            cross_corr_stationality_month={}
            for k in self.crosscorr_dataframe.keys():
                cross_corr_stationality_month[k]=self.crosscorr_dataframe[k][prii]
            statistics_observed_dataframe[prii]=self.statistics_dataframe[prii].mean(axis=1)
            #evaluator = evaluateInd_PSO(self.statististics_dataframe.loc[:,str(prii)].values, self.hiperparams.weights, self.hiperparams.process, self.hiperparams.statistics_name)
            evaluator = evaluateInd_PSO(self.statistics_dataframe[prii].mean(axis=1), self.hiperparams.weights, self.hiperparams.process, self.hiperparams.statistics_name, 
                                        self.hiperparams.storm_radius, cross_corr_stationality_month)
            if verbose:
                print('')
                print('Fitting the months = ' + str(prii))
                print('')
            error=0.5
            error_initializations=list()
            parameters_initializations=list()
            n_i=0
            total_error=10000
            while total_error>error and self.hiperparams.number_initializations>n_i:
                if verbose:
                    print('Initialization number  = ' + str(n_i))
                n_i+=1
                Steps = 0
                mySwarm = pso.Swarm(n_p, limites, self.hiperparams.number_bees, evaluator, 'min', verbose)
                
                while True:
                    if verbose:
                        print('Iteration number  = ' + str(Steps))
                    mySwarm.step()
                    Steps += 1
                    (b, bp) = mySwarm.getBest()
                    if verbose:
                        print('Total error = ' + str(evaluator.totalError(bp)))
                        if verbose:
                            if total_error <= error:
                                print('Stopped calibration since total error is equal or lower than error bound (' + str(error) + ')') 

                    if (evaluator.totalError(bp) < error): break
                    elif Steps > self.hiperparams.number_iterations: ###1000
                        break
                error_initializations.append(evaluator.totalError(bp))	
                parameters_initializations.append(bp)
                sta_fit=evaluator.compute(bp);
                #statististics_fit_dataframe.loc[:,str(prii)]= sta_fit[0]
                statistics_dataframe_fit.loc[:,str(prii)]=sta_fit[0][0:sta_len-nn_cross]
                for nn_ccc, st_cross in enumerate(self.hiperparams.statistics_name[sta_len-nn_cross:]):
                    crosscorr_dataframe_fit[st_cross][prii]['cross_corr'] = sta_fit[0][(sta_len-nn_cross)+(nn_ccc*cross_corr_division):(sta_len-nn_cross)+((nn_ccc+1)*cross_corr_division)]
                Fitted_parameters=mySwarm.getBest()[1]
                Error_dataframe.loc[str(prii)]=evaluator.totalError(bp)

                total_error=evaluator.totalError(bp)

            if n_i>1:
                # If in the first initialisation it does not find the best parameters, it does it again and keeps the simulation with the lowest error.
                Error_dataframe.loc[str(prii)]=np.min(error_initializations)
                sta_fit=evaluator.compute(parameters_initializations[np.argmin(error_initializations)]);
                #statististics_fit_dataframe.loc[:,str(prii)]= sta_fit[0]
                statistics_dataframe_fit.loc[:,str(prii)]=sta_fit[0][0:sta_len-nn_cross]
                for nn_ccc, st_cross in enumerate(self.hiperparams.statistics_name[sta_len:]):
                    crosscorr_dataframe_fit[st_cross][prii]['cross_corr'] = sta_fit[0][(sta_len-nn_cross)+(nn_ccc*cross_corr_division):(sta_len-nn_cross)+((nn_ccc+1)*cross_corr_division)]
                Fitted_parameters=parameters_initializations[np.argmin(error_initializations)]



            if self.hiperparams.process=='normal' and self.hiperparams.storm_radius==False:

                param_v=list(); 
                param_v.append(Fitted_parameters[0]);
                param_v.append(Fitted_parameters[1]); 
                param_v.append(Fitted_parameters[2]); 
                param_v.append(Fitted_parameters[3]); 
                param_v.append(Fitted_parameters[4]); 
                Dataframe_params[prii]=param_v

            if self.hiperparams.process=='normal' and self.hiperparams.storm_radius==True:

                param_v=list(); 
                param_v.append(Fitted_parameters[0]);
                param_v.append(Fitted_parameters[1]); 
                param_v.append(Fitted_parameters[2]); 
                param_v.append(Fitted_parameters[3]); 
                param_v.append(Fitted_parameters[4]); 
                param_v.append(Fitted_parameters[5])
                Dataframe_params[prii]=param_v


            elif self.hiperparams.process=='storms' and self.hiperparams.storm_radius==False:

                param_v=list();
                param_v.append(Fitted_parameters[0]);
                param_v.append(Fitted_parameters[1]); 
                param_v.append(Fitted_parameters[2]); 
                param_v.append(Fitted_parameters[3]); 
                param_v.append(Fitted_parameters[4]); 
                param_v.append(Fitted_parameters[5]); 
                param_v.append(Fitted_parameters[6]); 
                param_v.append(Fitted_parameters[7]); 
                param_v.append(Fitted_parameters[8]); 
                param_v.append(Fitted_parameters[9]);
                Dataframe_params[prii]=param_v
                
            elif self.hiperparams.process=='storms' and self.hiperparams.storm_radius==True:

                param_v=list();
                param_v.append(Fitted_parameters[0]);
                param_v.append(Fitted_parameters[1]); 
                param_v.append(Fitted_parameters[2]); 
                param_v.append(Fitted_parameters[3]); 
                param_v.append(Fitted_parameters[4]); 
                param_v.append(Fitted_parameters[5]); 
                param_v.append(Fitted_parameters[6]); 
                param_v.append(Fitted_parameters[7]); 
                param_v.append(Fitted_parameters[8]); 
                param_v.append(Fitted_parameters[9]); 
                param_v.append(Fitted_parameters[10]); 
                param_v.append(Fitted_parameters[11]); 
                Dataframe_params[prii]=param_v
        
        Dataframe_xi_months = XI_MONTHS(time_series, Dataframe_params, self.hiperparams.process) #calculate scale parameter for every gauge

        resuls = parameters_calibration(Dataframe_params,statistics_dataframe_fit, crosscorr_dataframe_fit, self.statistics_dataframe, self.crosscorr_dataframe,self.crosscorr_dataframe_dist, Error_dataframe, Dataframe_xi_months)
        
        
        
        
        return resuls
    



'''
Library containing the mathematical functions of the Neyman-Scott point model Rectangular Pulses (NSRP)

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
        
        
Point Model (NSRP)

	The NSRP process has the following parameters with their respective definitions:

	    landa --> Storms occur in a Poisson porcess of rate landa. Unidades: (time-1)
		mu_c  --> C number of cell origins is associated with each storm origin. An 
			  approach that ensures at least one rain cell follows each storm origin, 
			  is to take C-1 as a Poisson random variable with mean mu_c -1. 
			  Unidades: (numero de celldas por tormenta)
		eta   --> The duration of the pulse (rain cell) is an independent exponencial 
			  random variable with parameter eta. Unidades (time-1)
		beta  --> The cell origins being independently displaced from the storm origins
			  by distances that are exponetially distributed with parameter betha.
			  Unidades (time-1)
		xi    --> The intensity of the pulse can be exponencial, gamma or weibull 
			  distributed with parameteres escale pareameter xi. Unidades (mm-1)
		alpha --> the intensity of the pulse can be exponencial, gamma or weibull 
			  distributed with parameteres shape pareameter alpha. No tiene unidades.
'''


import numpy as np
import scipy as sp
import math as mt
from math import *


def E_X(xi, alpha, r):
    """Returns the intensity function (mm) that I am going to use (cell intensity).
       E' (exponential), 'W' (Weibull) and 'G' (Gamma).

    Exponential (1 parameters xi). The exponential function 
    is the same as the Weibull only with alpha=1. 
    Weibull (2 parameters, xi (scale parameter) and alpha (shape parameter).
    Gamma (2 parameters, xi (scale parameter) and alpha (shape parameter)
    
    xi:    Scale parameter
    alpha: Shape parameter
    r:     Exponent to which it is raised (e.g. E(X^r))"""
    
    E_X=(xi**(1*r))*mt.gamma(1+r*1)
    return E_X



def NSRP_mean(h, landa, mu_c, eta, xi, alpha, alpha_p):
    """Returns the average precipitation (mm/h) at a point.
    h : selected time interval (hours or days)"""

    result=list()
    for i in range(np.size(alpha_p)):
        result.append(h*landa[i]*mu_c[i]/(eta[i]*E_X(xi[i], alpha[i], 1)))
    return np.sum(result)

def NSRP_mean_ST(h, landa, mu_c, eta, xi, alpha, alpha_p):
    """Returns the average precipitation (mm/h) at a point.
    h : selected time interval (hours or days)"""

    result=list()
    for i in range(np.size(alpha_p)):
        result.append(h*landa[i]*mu_c[i]*E_X(xi[i], alpha[i], 1)/(eta[i]))
    return np.sum(result)


def NSRP_covariance(h,l, landa, mu_c, eta, xi, betha, alpha, alpha_p):
    """Returns the covariance.
    Eq (12)--> Further Developments of the Neyman-Scott Clustered Point 
    Process for Modeling Rainfall (Cowperwait, 1991)"""

    result=list()
    for i in range(np.size(landa)):
        if l==0:
            A=eta[i]*h-1+exp(-eta[i]*h)
            B=betha[i]*h-1+exp(-betha[i]*h)
        elif l>0:
            A=0.5*((1-exp(-eta[i]*h))**2)*exp(-eta[i]*h*(l-1))
            B=0.5*((1-exp(-betha[i]*h))**2)*exp(-betha[i]*h*(l-1))
        result.append(alpha_p[i]*(landa[i]*(eta[i]**-3)*A*\
        (2*mu_c[i]*E_X(1/xi[i], alpha[i], 2) +\
	(E_X( 1/xi[i], alpha[i], 1)**2)*(betha [i]**2)*\
	(mu_c[i]**2)/((betha[i]**2)-(eta[i]**2)))-\
        landa[i]*(E_X(1/xi[i], alpha[i], 1)**2)*B*(mu_c[i]**2)/\
	(betha[i]*((betha[i]**2)-(eta[i]**2)))))
    return np.sum(result)
	

def multiply(numbers):  
    total = 1
    for x in numbers:
        total *= x  
    return total  


def NSRP_pdry(h, landa, mu_c, eta, betha, alpha_p):
    """Returns the probability that it does not rain in an arbitrary time of length h.
    Eq (6) Stochastic point porcess modelling of rainfall. I. 
    sigle-site fitting and validation (Cowperwait, 1995)
    We use the approximation to avoid having to solve the integral--> Further Developments of the Neyman-Scott Clustered Point 
    Process for Modeling Rainfall (Cowperwait, 1991)"""

    result=list()
    for i in range(np.size(alpha_p)):

        result.append((exp(-landa[i]*h + (landa[i]*(betha[i]**-1)*((mu_c[i]-1)**-1))*(1-exp(1 -mu_c[i]+(mu_c[i]-1)*exp(-betha[i]*h)))-\
        (landa[i]/betha[i])*(0.577 + np.log(((eta[i]/(eta[i]-betha[i]))-exp(-betha[i]*h))*(mu_c[i]))))))
    return multiply(result)

def moving_average(a, n) :
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n
    
    
def fi_h(Datos, h):
    return np.sum(moving_average(Datos.values, h)<=0.1)/(np.sum(Datos.values>=0))


def NSRP_fi_DD(h, landa, ipsilon, eta, betha, alpha_p):
    """Returns the probability that it will not rain in an arbitrary time of 
    length h conditional on the following time h not raining either.
    Cowperwait , 1995. Stochastic point process modelling of rainfall. I. 
    single-site fitting and validation. NSRP"""
    result=NSRP_pdry(2*h, landa, ipsilon, eta, betha, alpha_p)/\
        NSRP_pdry(h, landa, ipsilon, eta, betha, alpha_p)
    return result

def NSRP_fi_WW(h, landa, ipsilon, eta, betha, alpha_p):
    """Returns the probability that it will rain in an arbitrary time of 
    length h conditional on the following time h also being rained on."""
    result=(1-2*NSRP_pdry(h, landa, ipsilon, eta, betha, alpha_p) +\
                       NSRP_pdry(2*h, landa, ipsilon, eta, betha, alpha_p))/\
                       (1-NSRP_pdry(h, landa, ipsilon, eta, betha, alpha_p))
    return result


def fi_DD(Datos, h):
    """(7, 8, 9). Cowperwait , 1995. Stochastic point process modelling of rainfall. I. 
    single-site fitting and validation. NSRP"""
    return fi_h(Datos, 2*h)/fi_h(Datos, h)
    
    
def fi_WW(Datos, h):
    """(7, 8, 9). Cowperwait , 1995. Stochastic point process modelling of rainfall. I. 
    single-site fitting and validation. NSRP"""
    return (1 - 2*fi_h(Datos, h)+fi_h(Datos, 2*h))/(1-fi_h(Datos, h))


def f_eta_beta_h(neta, beta, h):
    f=-2*neta**3*beta**2*exp(-neta*h)-2*neta**3*beta**2*exp(-beta*h)+neta**2*beta**3*exp(-2*neta*h)+2*neta**4*beta*exp(-neta*h)+2*neta**4*beta*exp(-beta*h)\
    +2*neta**3*beta**2*exp(-(neta+beta)*h)-2*neta**4*beta*exp(-(neta+beta)*h)-8*neta**3*beta**3*h+11*neta**2*beta**3-2*neta**4*beta\
    +2*neta**3*beta**2+4*neta*beta**5*h+4*neta**5*beta*h-7*beta**5-4*neta**5+8*beta**5*exp(-neta*h)-beta**5*exp(-2*neta*h)\
    -2*h*neta**3*beta**3*exp(-neta*h)-12*neta**2*beta**3*exp(-neta*h)+2*h*neta*beta**5*exp(-neta*h)+4*neta**5*exp(-beta*h)
    return f
def g_eta_beta_h(neta, beta, h):
    g=12*neta**5*beta*exp(-beta*h)+9*neta**4*beta**2+12*neta*beta**5*exp(-neta*h)+9*neta**2*beta**4+12*neta**3*beta**3*exp(-(neta+beta)*h)\
    -neta**2*beta**4*exp(-2*neta*h)-12*neta**3*beta**3*exp(-beta*h)-9*neta**5*beta-9*neta*beta**5-3*neta*beta**5*exp(-2*neta*h)\
    -neta**4*beta**2*exp(-2*beta*h)-12*neta**3*beta**3*exp(-neta*h)+6*neta**5*beta**2*h-10*beta**4*neta**3*h+6*beta**5*neta**2*h\
    -10*beta**3*neta**4*h+4*beta**6*neta*h-8*beta**2*neta**4*exp(-beta*h)+4*beta*neta**6*h+12*beta**3*neta**3\
    -8*beta**4*neta**2*exp(-neta*h)-6*neta**6-6*beta**6-2*neta**6*exp(-2*beta*h)-2*beta**6*exp(-2*neta*h)\
    +8*neta**6*exp(-beta*h)+8*beta**6*exp(-neta*h)-3*beta*neta**5*exp(-2*beta*h)
    return g

def NSRP_moments_order_3(C,h,l, landa, mu_c, eta, xi, betha, alpha, alpha_p):
    """Returns the third-order moments.
    Eq (12)--> Further Developments of the Neyman-Scott Clustered Point 
    Process for Modeling Rainfall (Cowperwait, 1991)"""

    result=list()
    for i in range(np.size(landa)):
        if l==0:
            A=eta[i]*h-1+exp(-eta[i]*h)
            B=betha[i]*h-1+exp(-betha[i]*h)
        elif l>0:
            A=0.5*((1-exp(-eta[i]*h))**2)*exp(-eta[i]*h*(l-1))
            B=0.5*((1-exp(-betha[i]*h))**2)*exp(-betha[i]*h*(l-1))
	
        mu_x = E_X( 1/xi[i], alpha[i], 1)
        Ex2  = E_X( 1/xi[i], alpha[i], 2) 
        Ex3  = E_X( 1/xi[i], alpha[i], 3)
        
        if C=='Poisson':
            E_C1=mu_c[i]**2; E_C2=mu_c[i]**3
        if C=='Geometric':
            E_C1=2*mu_c[i]*(mu_c[i]-1); E_C2=6*mu_c[i]*(mu_c[i]-1)**2
        
        result.append(((6*landa[i]*mu_c[i]*Ex3*(eta[i]*h-2+eta[i]*h*exp(-eta[i]*h)\
		+2*exp(-eta[i]*h))/(eta[i]**4))+(3*landa[i]*mu_x*Ex2*(E_C1)*\
		f_eta_beta_h(eta[i], betha[i],h))/(2*(eta[i]**4)*betha[i]*\
		((betha[i]**2)-(eta[i]**2))**2))+landa[i]*(mu_x**3)*(E_C2)*\
		g_eta_beta_h(eta[i], betha[i],h)/(2*(eta[i]**4)*betha[i]*((eta[i]**2)-\
		(betha[i]**2))*(eta[i]-betha[i])*(2*betha[i]+eta[i])*(betha[i]+2*eta[i])))
    return np.sum(result)





'''
Library containing classes for keeping the simulation output.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''

import pandas as pd

class outputs_simulation (object):
    def __init__(self,df_sim_join_daily,df_sim_join_hour,statististics_sim_df,temporal_resolution):
        self.Daily_Simulation        = df_sim_join_daily
        self.Hourly_Simulation       = df_sim_join_hour
        self.statististics_Simulated = statististics_sim_df
        self.temporal_resolution = temporal_resolution

    def save_files(self,path_output_files):
        self.Daily_Simulation.to_csv(path_output_files+'Time_serie_daily_simulated.csv')
        self.Hourly_Simulation.to_csv(path_output_files+'Time_serie_hourly_simulated.csv')
        if self.temporal_resolution == 'd':
            self.statististics_Simulated.to_csv(path_output_files+'statistic_daily_simulated.csv')
        elif self.temporal_resolution == 'h':
            self.statististics_Simulated.to_csv(path_output_files+'statistic_hourly_simulated.csv')



'''
Library containing classes for simulating time series from the model.
	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''



from NEOPRENE.STNSRP.MathematicalPropertiesSTNSRP import *
from NEOPRENE.STNSRP.utils import *
from NEOPRENE.STNSRP.libs_STNSRP import *
from NEOPRENE.STNSRP.inputs_simulation import *
from NEOPRENE.STNSRP.outputs_simulation import *
from NEOPRENE.STNSRP.Statistics import Statistics
import time

class Simulation(object):
    def __init__(self,hiperparams):
        self.hiperparams = hiperparams
        
    def __call__(self, params_cal, Input_Series, Input_Attr):
   
        statististics_sim_df=pd.DataFrame(index=self.hiperparams.statistics_name,columns=self.hiperparams.Seasonality_str)
    
        list_params = inputs_simulation(params_cal).params_cal
        
        Df_params = list_params[0]
        
        Df_params.columns = self.hiperparams.Seasonality
        
        Dataframe_xi_months = list_params[1]
        
        Df_params=allmonths(Df_params)
        
        XX = Input_Attr.X
        YY = Input_Attr.Y
        
        print('')
        print('')
        print('#'*80)
        print('Synthetic simulation')
        print('')
        print('')


		## We start with synthetic simulation and statistical calculations.

        if self.hiperparams.process=='normal':
   
            [Df_sim_join_day_,Df_sim_join_hour_]=\
                STNSRP_simulation(Df_params, Dataframe_xi_months, XX, YY, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, self.hiperparams.process,
                                 self.hiperparams.coordinates,self.hiperparams.storm_radius, self.hiperparams.Seasonality, Input_Attr.ID.values)

        elif self.hiperparams.process=='storms':

            if self.hiperparams.storm_radius == False:

                Df_params1=pd.DataFrame(index=['landa', 'ipsilon', 'eta', 'betha', 'fi_may'],columns=Df_params.columns)
                for i in Df_params.columns:
                    Df_params1.loc[:,i]=Df_params.loc[:,i].values[0:5]

                Df_params2=pd.DataFrame(index=['landa', 'ipsilon', 'eta', 'betha', 'fi_may'],columns=Df_params.columns)
                for i in Df_params.columns:
                    Df_params2.loc[:,i]=Df_params.loc[:,i].values[5:]

            elif self.hiperparams.storm_radius == True:

                Df_params1=pd.DataFrame(index=['landa', 'ipsilon', 'eta', 'betha', 'fi_may', 'fi_may_s'],columns=Df_params.columns)
                for i in Df_params.columns:
                    Df_params1.loc[:,i]=Df_params.loc[:,i].values[0:6]

                Df_params2=pd.DataFrame(index=['landa', 'ipsilon', 'eta', 'betha', 'fi_may', 'fi_may_s'],columns=Df_params.columns)
                for i in Df_params.columns:
                    Df_params2.loc[:,i]=Df_params.loc[:,i].values[6:]


            [Df_sim_join_day_1,Df_sim_join_hour_1]=\
            STNSRP_simulation(Df_params1, Dataframe_xi_months, XX, YY, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, 'normal',
                                 self.hiperparams.coordinates,self.hiperparams.storm_radius, self.hiperparams.Seasonality, Input_Attr.ID.values)

            [Df_sim_join_day_2,Df_sim_join_hour_2]=\
            STNSRP_simulation(Df_params2, Dataframe_xi_months, XX, YY, self.hiperparams.year_ini, self.hiperparams.year_fin, self.hiperparams.temporal_resolution, 'normal',
                                 self.hiperparams.coordinates,self.hiperparams.storm_radius, self.hiperparams.Seasonality, Input_Attr.ID.values)


            Df_sim_join_hour_=Df_sim_join_hour_1 + Df_sim_join_hour_2
            Df_sim_join_day_=Df_sim_join_day_1 + Df_sim_join_day_2 
        
        ## Real, adjusted and simulated statistical calculations.
        

        if self.hiperparams.temporal_resolution == 'd':
            Data=Df_sim_join_day_.copy()
        elif self.hiperparams.temporal_resolution == 'h':
            Data=Df_sim_join_hour_.copy()
            
        Data[Data<0.001]=0
    
        statististics_sim_df = Statistics(self.hiperparams, time_series = Data, attributes = Input_Attr)
        
        statistics_dataframe        = statististics_sim_df.statistics_dataframe.copy()
        crosscorr_dataframe         = statististics_sim_df.crosscorr_dataframe.copy()
        crosscorr_dataframe_dist    = statististics_sim_df.crosscorr_dist_dataframe.copy()

        results = outputs_simulation(Df_sim_join_day_,Df_sim_join_hour_,statistics_dataframe,crosscorr_dataframe,crosscorr_dataframe_dist, self.hiperparams.temporal_resolution,self.hiperparams.Seasonality)

        return results



'''
Library containing classes for keeping the simulated model parameters.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''


import pandas as pd

class parameters_calibration(object):
    def __init__(self, Dataframe_parametros, statististics_fit_dataframe, crosscorr_dataframe_fit, statististics_Real, crosscorr_Real,crosscorr_dataframe_dist, Error_dataframe, Dataframe_xi_months):
        self.Fitted_parameters  = Dataframe_parametros
        self.statististics_Fit  = statististics_fit_dataframe
        self.statististics_Real = statististics_Real
        self.crosscorr_Fit  = crosscorr_dataframe_fit
        self.crosscorr_Real = crosscorr_Real
        self.crosscorr_Real_Dist = crosscorr_dataframe_dist
        self.Error_dataframe    = Error_dataframe
        self.Dataframe_xi_months    = Dataframe_xi_months

    def save_files(self, path_output_files): 
        self.Fitted_parameters.to_csv(path_output_files   + 'Calibrated_parameters.csv')
        self.statististics_Fit.to_csv(path_output_files   + 'statististics_fit.csv')
        self.Dataframe_xi_months.to_csv(path_output_files + 'xi_months.csv')
        
        list_crosscorr = list()
        for j in self.crosscorr_Real.keys():
            if 'crosscorr_' in j:
                list_crosscorr.append(j)
        
        
        for i in self.Fitted_parameters.columns:
            self.statististics_Real[i].to_csv(path_output_files+'statististics_real_'+str(i)+'.csv')
            if len(list_crosscorr)!=0:
                for c in list_crosscorr:
                    self.crosscorr_Real[c][i].to_csv(path_output_files+c+'_real_'+str(i)+'.csv')
                    self.crosscorr_Fit[c][i].to_csv(path_output_files +c+'_Fit_'+str(i)+'.csv')
            
            
            
        


'''
Library to perform Particle Swarm Optimization.

	Authors: 
        + Manuel del Jesus
    
'''

import numpy as np

class Worker(object):
    __slots__ = ["_position", "_velocity", "_best", "_best_position", "_parent", "fvalue"]
    OMEGA = 0.9
    PHIW = 0.9
    PHIS = 0.9
    def __init__(self, parent):
        self._position = np.random.rand(parent._D) * parent._inc + parent._inf
        self._velocity = np.random.rand(parent._D) * parent._inc
        self._best = parent._Function(self._position)
        self._best_position = self._position.copy()
        self._parent = parent
        self.fvalue = None

    def update(self, swarmBest):
        self._position += self._velocity
        self._position = np.maximum(np.minimum(self._position, self._parent._limits[:,1]),self._parent._limits[:,0])
        #print "++ ", swarmBest, self._best_position, self._position
        self._velocity = self.OMEGA * self._velocity \
            + self.PHIW * np.random.rand(self._parent._D) * (self._best_position - self._position) \
            + self.PHIS * np.random.rand(self._parent._D) * (swarmBest - self._position)
        #self._velocity /= np.sqrt(np.sum(self._velocity**2))
        #print np.sqrt(np.sum(self._velocity**2))
        #print self._position, swarmBest
        self.fvalue = self._parent._Function(self._position)
        self.testValue()

    def testValue(self):
        raise NotImplementedError("You are calling the parent class.")


class Worker_min(Worker):
    def testValue(self):
        if self.fvalue < self._best:
            #In case the current value is better than the old one
            self._best = self.fvalue
            self._best_position = self._position.copy()
        self.fvalue = None

class Worker_max(Worker):
    def testValue(self):
        if self.fvalue > self._best:
            #In case the current value is better than the old one
            self._best = self.fvalue
            self._best_position = self._position.copy()
        self.fvalue = None

class Swarm(object):
    def __init__(self, D, limits, n, F, optimum, verbose=False):
        self._D = D
        self._limits = limits
        self._inf = limits[:,0]
        self._inc = limits[:,1] - limits[:,0]
        self._Function = F
        if optimum=='min':
            self._workers = [Worker_min(self) for i in range(n)]
        elif optimum=='max':
            self._workers = [Worker_max(self) for i in range(n)]
        else:
            raise SystemError("Optimum not known...\toptimum = %s"%optimum)
        self._bestList = [i._best for i in self._workers]
        if optimum=='min':
            self._best = np.min(self._bestList)
            self._best_position = self._workers[np.argmin(self._bestList)]._best_position.copy()
        elif optimum=='max':
            self._best = np.max(self._bestList)
            self._best_position = self._workers[np.argmax(self._bestList)]._best_position.copy()
        if verbose:
            print("Initial best position ", self._best, self._best_position)

    def getBest(self):
        #print("Best value : %6.5e"%self._best, end=" ")
        #print("Position : ", end=" ")
        #print("\t", self._best_position)
        return (self._best, self._best_position)

    def setBest(self, posicion):
        best = self._Function(posicion)
        try:
            if best < self._best:
                self._best = best
                self._best_position = posicion
        except:
            self._best = best
            self._best_position = posicion

    def step(self):
        [i.update(self._best_position) for i in self._workers]
        self._bestList = [i._best for i in self._workers]
        tentative = np.min(self._bestList)
        if tentative < self._best:
            self._best = tentative
            self._best_position = self._workers[np.argmin(self._bestList)]._best_position.copy()

def testFunction(x):
    return np.sum((x-0.83)**2)
    #return (1-x[0])**2 + 100.0*(x[1]-x[0]**2)**2
    #return np.sum(100.0*(x[1:] - x[:-1]**2)**2 + (x[:-1] - 1.0)**2)

if __name__ == "__main__":
    a = 0.8
    b = 1.2
    S = Swarm(6, np.asarray([[a,b],[a,b],[a,b],[a,b],[a,b],[a,b]]), 100, testFunction, 'min')

    with open("results_prueba.txt",'wt', 0) as fid:
        for i in xrange(100):
            for j in xrange(100):
                S.step()
            (b, bp) = S.getBest()
            fid.write('%6.5e\t%s\n'%(b, str(bp)))





'''
Library containing the main functions for the execution of the model.

	Authors: 
        + Javier Díez Sierra
	    + Salvador Navas Fernández
        + Manuel del Jesus
    
'''

import pandas as pd

class inputs_simulation (object):
    def __init__(self,inputs_sim):
            
        if type(inputs_sim) != str and type(inputs_sim)!= list:
            self.params_cal = [inputs_sim.Fitted_parameters,inputs_sim.Dataframe_xi_months]
        elif type(inputs_sim) == str or type(inputs_sim) == list:
            self.params_cal = [pd.read_csv(inputs_sim[0],index_col=0), pd.read_csv(inputs_sim[1],index_col=0)]
        else:
            raise Exception


from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
#long_description = (this_directory / "README.md").read_text()
with open(this_directory / "README.md", encoding="utf-8") as f:
    long_description = f.read()


setup(  name='NEOPRENE',
        packages = find_packages(),
        license = "GPLv3",
        version='0.99.1',
        description='🌎 Scripts and information to synthetic generation of precipitation based on Point Processes.',
        long_description=long_description,
        long_description_content_type='text/markdown',
        author='Javier Diez Sierra <javier.diez@unican.es>, Salvador Navas <salvador.navas@unican.es>, Manuel del Jesus <manuel.deljesus@unican.es>',
        author_email='javier.diez@unican.es, salvador.navas@unican.es, manuel.deljesus@unican.es',
        maintainer       = 'Manuel del Jesus',
        maintainer_email = 'manuel.deljesus@unican.es',
        url = 'https://github.com/IHCantabria/NEOPRENE',
        include_package_data=True,
        #package_data={'': ['notebooks/*.ipynb']},
        #data_files=[("notebooks",["*"])],
        python_requires='>=3.7, <4',
        install_requires=[
            'numpy',
            'pandas',
            'scipy',
            'datetime',
            'matplotlib',
            'pyyaml',
            'haversine',
            'tqdm',
            'pyshp',
            'shapely',
        ],
        extras_require={'plotting': ['matplotlib>=2.2.0', 'jupyter','jupyterlab']}
        )
